#!/usr/bin/env python
# -*- coding: utf-8 -*-
# #########################################################################
# Copyright (c) 2020, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2020. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# #########################################################################

import sys, numpy, os

from PyQt5.QtWidgets import QMessageBox

from orangewidget import gui
from orangewidget.widget import OWAction
from orangewidget.settings import Setting

from oasys.widgets.exchange import DataExchangeObject
from oasys.widgets import gui as oasysgui
from oasys.widgets import widget, congruence

class TotalFilterCalculator(widget.OWWidget):

    name = "Total Filter Calculator"
    description = "Total Filter Calculator"
    icon = "icons/total_filter.png"
    maintainer = "Luca Rebuffi"
    maintainer_email = "lrebuffi(@at@)anl.gov"
    priority = 5.01
    category = "User Defined"
    keywords = ["data", "file", "load", "read"]

    inputs = [("Mirror Reflectivity #1", DataExchangeObject, "set_ref_1"),
              ("Mirror Reflectivity #2", DataExchangeObject, "set_ref_2"),
              ("Mirror Reflectivity #3", DataExchangeObject, "set_ref_3"),
              ("Diffraction Profile #1", DataExchangeObject, "set_dif_1"),
              ("Diffraction Profile #2", DataExchangeObject, "set_dif_2"),
              ("MultiLayer Reflectivity #1", DataExchangeObject, "set_mul_1"),
              ("MultiLayer Reflectivity #2", DataExchangeObject, "set_mul_2"),
              ("Power Output #1", DataExchangeObject, "set_pow_1")]

    outputs = [{"name":"Total Filter",
                "type":DataExchangeObject,
                "doc":"Total Filter",
                "id":"total_filter"}]

    want_main_area = 1

    ref_1 = None
    ref_2 = None
    ref_3 = None
    dif_1 = None
    dif_2 = None
    mul_1 = None
    mul_2 = None
    pow_1 = None

    check_ref_1 = Setting(0)
    check_ref_2 = Setting(0)
    check_ref_3 = Setting(0)
    check_dif_1 = Setting(0)
    check_dif_2 = Setting(0)
    check_mul_1 = Setting(0)
    check_mul_2 = Setting(0)
    check_pow_1 = Setting(0)

    n_ref_1 = Setting(0)
    n_ref_2 = Setting(0)
    n_ref_3 = Setting(0)
    n_dif_1 = Setting(0)
    n_dif_2 = Setting(0)
    n_mul_1 = Setting(0)
    n_mul_2 = Setting(0)

    energy_from = Setting(1000.0)
    energy_to   = Setting(51000.0)
    energy_nr = Setting(5000)

    CONTROL_AREA_WIDTH = 405

    def __init__(self):
        super(TotalFilterCalculator, self).__init__()

        self.runaction = OWAction("Calculate Total Filter", self)
        self.runaction.triggered.connect(self.calculate_total_filter)
        self.addAction(self.runaction)

        self.setFixedWidth(1200)
        self.setFixedHeight(710)

        self.controlArea.setFixedWidth(self.CONTROL_AREA_WIDTH)

        box0 = oasysgui.widgetBox(self.controlArea, "", orientation="horizontal", width=self.CONTROL_AREA_WIDTH-5)
        gui.button(box0, self, "Calculate Total Filter", callback=self.calculate_total_filter, height=45)

        box1 = oasysgui.widgetBox(self.controlArea, "Energy Range", orientation="vertical", width=self.CONTROL_AREA_WIDTH-5)

        oasysgui.lineEdit(box1, self, "energy_from", "Energy From [eV]", labelWidth=250, orientation="horizontal", valueType=float)
        oasysgui.lineEdit(box1, self, "energy_to", "Energy to [eV]", labelWidth=250, orientation="horizontal", valueType=float)
        oasysgui.lineEdit(box1, self, "energy_nr", "Nr. of energies", labelWidth=250, orientation="horizontal", valueType=int)

        box1 = oasysgui.widgetBox(self.controlArea, "Input Data", orientation="vertical", width=self.CONTROL_AREA_WIDTH-5)

        items = ["1", "2", "3", "4", "5", "6", "7", "8", "9", "10"]

        def build_combo(container, master, field_name, label, has_combo=True):
            box = gui.widgetBox(container, "", orientation="horizontal")
            gui.checkBox(box, master, "check_" + field_name[2:], label="" if has_combo else label)
            if has_combo: gui.comboBox(box, master, field_name, label=label, items=items, labelWidth=350, sendSelectedValue=False, orientation="horizontal")
            box.setEnabled(False)

            return box

        self.cb_ref_1 = build_combo(box1, self, "n_ref_1", label="Nr. of Mirror Reflectivity #1")
        self.cb_ref_2 = build_combo(box1, self, "n_ref_2", label="Nr. of Mirror Reflectivity #2")
        self.cb_ref_3 = build_combo(box1, self, "n_ref_3", label="Nr. of Mirror Reflectivity #3")

        gui.separator(box1)

        self.cb_dif_1 = build_combo(box1, self, "n_dif_1", label="Nr. of Diffraction Profile #1")
        self.cb_dif_2 = build_combo(box1, self, "n_dif_2", label="Nr. of Diffraction Profile #2")

        gui.separator(box1)

        self.cb_mul_1 = build_combo(box1, self, "n_mul_1", label="Nr. of MultiLayer Reflectivity #1")
        self.cb_mul_2 = build_combo(box1, self, "n_mul_2", label="Nr. of MultiLayer Reflectivity #2")

        gui.separator(box1)

        self.cb_pow_1 = build_combo(box1, self, "n_pow_1", label="Power Widget Input", has_combo=False)

        gui.rubber(self.controlArea)

        tabs = oasysgui.tabWidget(self.mainArea)
        tabs.setFixedHeight(self.height()-15)
        tabs.setFixedWidth(770)

        tab_fil = oasysgui.createTabPage(tabs, "Total Filter")

        self.filter_plot = oasysgui.plotWindow(tab_fil, position=True)
        self.filter_plot.setFixedHeight(self.height()-60)
        self.filter_plot.setFixedWidth(760)
        self.filter_plot.setGraphXLabel("Energy [eV]")
        self.filter_plot.setGraphYLabel("Intensity Factor")
        self.filter_plot.setGraphTitle("Total Filter")


    def __get_mirror_reflectivity(self, exchange_data):
        try:
            if not exchange_data.get_program_name() == "XOPPY": raise ValueError("Only XOPPY widgets are accepted")
            if not exchange_data.get_widget_name() == "XF1F2": raise ValueError("Only XF1F2 widgets are accepted")

            return exchange_data.get_content("xoppy_data")
        except Exception as e:
            QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)

            if self.IS_DEVELOP: raise e

    def __get_diffraction_profile(self, exchange_data):
        try:
            if not exchange_data.get_program_name() == "XOPPY": raise ValueError("Only XOPPY widgets are accepted")
            if not exchange_data.get_widget_name() == "XCRYSTAL": raise ValueError("Only XCRYSTAL widgets are accepted")

            xoppy_data = exchange_data.get_content("xoppy_data")

            energies = xoppy_data[:, 0]

            reflectivity_s = xoppy_data[:, 5]
            reflectivity_p = xoppy_data[:, 6]

            reflectivity_s[numpy.where(numpy.isnan(reflectivity_s))] = 0.0
            reflectivity_p[numpy.where(numpy.isnan(reflectivity_p))] = 0.0

            data = numpy.zeros((len(energies), 2))
            data[:, 0] = energies
            data[:, 1] = 0.5 * (reflectivity_p + reflectivity_s)

            return data
        except Exception as e:
            QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)

            if self.IS_DEVELOP: raise e

    def __get_multilayer_reflectivity(self, exchange_data):
        try:
            if not exchange_data.get_program_name() == "XOPPY": raise ValueError("Only XOPPY widgets are accepted")
            if not exchange_data.get_widget_name() == "MULTILAYER": raise ValueError("Only MULTILAYER widgets are accepted")

            xoppy_data = exchange_data.get_content("xoppy_data")

            energies = xoppy_data[:, 0]

            reflectivity_s = xoppy_data[:, 1]
            reflectivity_p = xoppy_data[:, 2]

            reflectivity_s[numpy.where(numpy.isnan(reflectivity_s))] = 0
            reflectivity_p[numpy.where(numpy.isnan(reflectivity_p))] = 0

            data = numpy.zeros((len(energies), 2))
            data[:, 0] = energies
            data[:, 1] = 0.5 * (reflectivity_p + reflectivity_s)

            return data
        except Exception as e:
            QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)

            if self.IS_DEVELOP: raise e

    def __get_power_data(self, exchange_data):
        try:
            if not exchange_data.get_program_name() == "XOPPY": raise ValueError("Only XOPPY widgets are accepted")
            if not exchange_data.get_widget_name() == "POWER": raise ValueError("Only POWER widgets are accepted")

            xoppy_data = exchange_data.get_content("xoppy_data")

            energies = xoppy_data[:, 0]

            data = numpy.zeros((len(energies), 2))
            data[:, 0] = energies
            data[:, 1] = xoppy_data[:, -1] / xoppy_data[:, 1]

            return data
        except Exception as e:
            QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)

            if self.IS_DEVELOP: raise e

    def __set_data(self, exchange_data, getter_method, field, combobox):
        if exchange_data is None: 
            setattr(self, field, None)
            setattr(self, "check_" + field, 0)
            combobox.setEnabled(False)
        else:
            try:
                setattr(self, field, getter_method(exchange_data))
                setattr(self, "check_" + field, 1)
                combobox.setEnabled(True)
            except Exception as e:
                combobox.setEnabled(False)
                
                QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)
                
                if self.IS_DEVELOP: raise e

    def set_ref_1(self, exchange_data): self.__set_data(exchange_data, self.__get_mirror_reflectivity,     "ref_1", self.cb_ref_1)
    def set_ref_2(self, exchange_data): self.__set_data(exchange_data, self.__get_mirror_reflectivity,     "ref_2", self.cb_ref_2)
    def set_ref_3(self, exchange_data): self.__set_data(exchange_data, self.__get_mirror_reflectivity,     "ref_3", self.cb_ref_3)
    def set_dif_1(self, exchange_data): self.__set_data(exchange_data, self.__get_diffraction_profile,     "dif_1", self.cb_dif_1)
    def set_dif_2(self, exchange_data): self.__set_data(exchange_data, self.__get_diffraction_profile,     "dif_2", self.cb_dif_2)
    def set_mul_1(self, exchange_data): self.__set_data(exchange_data, self.__get_multilayer_reflectivity, "mul_1", self.cb_mul_1)
    def set_mul_2(self, exchange_data): self.__set_data(exchange_data, self.__get_multilayer_reflectivity, "mul_2", self.cb_mul_2)
    def set_pow_1(self, exchange_data): self.__set_data(exchange_data, self.__get_power_data,              "pow_1", self.cb_pow_1)

    def __add_data_to_total_filter(self, total_reflectivity, data, nr_data, check):
        if not data is None and check==1: total_reflectivity[:, 1] *= numpy.interp(total_reflectivity[:, 0], data[:, 0], data[:, 1])**(nr_data+1)

    def calculate_total_filter(self):
        try:
            congruence.checkStrictlyPositiveNumber(self.energy_from, "Energy From")
            congruence.checkStrictlyPositiveNumber(self.energy_to, "Energy to")
            congruence.checkStrictlyPositiveNumber(self.energy_nr, "Nr. of energies")
            congruence.checkLessThan(self.energy_from, self.energy_to, "Energy From", "Energy to")

            energies = numpy.linspace(self.energy_from, self.energy_to, self.energy_nr)

            total_filter = numpy.ones((self.energy_nr, 2))
            total_filter[:, 0] = energies

            self.__add_data_to_total_filter(total_filter, self.ref_1, self.n_ref_1, self.check_ref_1)
            self.__add_data_to_total_filter(total_filter, self.ref_2, self.n_ref_2, self.check_ref_2)
            self.__add_data_to_total_filter(total_filter, self.ref_3, self.n_ref_3, self.check_ref_3)

            self.__add_data_to_total_filter(total_filter, self.dif_1, self.n_dif_1, self.check_dif_1)
            self.__add_data_to_total_filter(total_filter, self.dif_2, self.n_dif_2, self.check_dif_2)

            self.__add_data_to_total_filter(total_filter, self.mul_1, self.n_mul_1, self.check_mul_1)
            self.__add_data_to_total_filter(total_filter, self.mul_2, self.n_mul_2, self.check_mul_2)

            self.__add_data_to_total_filter(total_filter, self.pow_1, 1, self.check_pow_1)

            total_filter[numpy.where(numpy.isnan(total_filter))] = 0.0

            if not total_filter[:, 1].sum() == total_filter.shape[0]:
                self.filter_plot.clear()
                self.filter_plot.addCurve(total_filter[:, 0], total_filter[:, 1], replace=True, legend="Total Filter")
                self.filter_plot.setGraphXLabel("Energy [eV]")
                self.filter_plot.setGraphYLabel("Intensity Factor")
                self.filter_plot.setGraphTitle("Total Filter")

                calculated_data = DataExchangeObject("ShadowOui_Thermal", "TOTAL_FILTER")
                calculated_data.add_content("total_filter", total_filter)

                self.send("Total Filter", calculated_data)
            else:
                raise ValueError("Calculation not possibile: no input data")
        except Exception as e:
            QMessageBox.critical(self, "Error", str(e), QMessageBox.Ok)

            if self.IS_DEVELOP: raise e






# #########################################################################
# Copyright (c) 2020, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2020. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# #########################################################################
#
# This is a translation from the matlab code:
#
#%% Function to perform the 0th Order Hankel Transform
#% Implemented by Joan Vila-Comamala from a routine based on the paper:
#%
#% M. Guizar-Sicairos and J. C. Gutierrez-Vega, Computation of quasi-discrete
#% Hankel transforms of integer order for propagating optical wave fields,
#% J. Opt. Soc. Am. A 21, 53-58 (2004).
#%
#% November 2010
#%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%

import numpy
from scipy.special import jv as besselj

#
# h is the 1D wavefield
# R is the maximum radius in Q space
# c are the zeros of the bessel functions

def Hankel_Transform_MGS(h, R, c, Nzeros=0):
    N = len(h)

    if Nzeros <= 0: Nzeros = N
    H = numpy.full(N, 0j)

    V = c[N]/(2*numpy.pi*R)

    h = h.conjugate().T

    m1 = (numpy.abs(besselj(1, c[0:N]))/R).conjugate().T
    m2 = m1*R/V

    F = numpy.divide(h, m1)

    Bessel_Jn = numpy.abs(besselj(1, c[0:Nzeros])) / (2 / c[N])
    Bessel_Jm = numpy.abs(besselj(1, c))

    for jj in range(0, N):
        C     = besselj(0, c[0:Nzeros] * c[jj] / c[N]) / (Bessel_Jn * Bessel_Jm[jj])
        H[jj] = numpy.dot(C[0:Nzeros], F[0:Nzeros])

    H = H.conjugate().T
    H = numpy.multiply(H, m2)
    H = H.conjugate().T

    return H


'''
N = size(h,2);

H(1:N) = 0;

if nargin < 4 || isempty(Nzero)
   Nzero=N;
end

V = c(N+1)/(2*pi*R);    % Maximum frequency

h=h';

m1 = (abs(besselj(1,c(1:N)))/R)';   %% m1 prepares input vector for transformation
m2 = m1*R/V;                        %% m2 prepares output vector for display

F = h./m1;
H(1:N)=0;

Bessel_Jn = abs(besselj(1,c(1:Nzero)))/(2/c(N+1));
Bessel_Jm = abs(besselj(1,c));

parfor jj=1:N
    C = besselj(0,c(1:Nzero)*c(jj)/c(N+1))./(Bessel_Jn*Bessel_Jm(jj));
    H(jj)=C(1:Nzero)*F(1:Nzero);
end

H = H';
H = H.*m2;
H = H';
'''



import sys, numpy, copy

from PyQt5.QtGui import QPalette, QColor, QFont
from PyQt5.QtWidgets import QMessageBox

from orangewidget import gui, widget
from orangewidget.settings import Setting
from oasys.widgets import gui as oasysgui
from oasys.widgets import congruence
from oasys.util.oasys_util import EmittingStream, TTYGrabber, TriggerIn

from srxraylib.util.inverse_method_sampler import Sampler2D

from orangecontrib.shadow.util.shadow_objects import ShadowOpticalElement, ShadowCompoundOpticalElement, ShadowBeam
from orangecontrib.shadow.util.shadow_util import ShadowCongruence, ShadowPhysics, ShadowMath
from orangecontrib.shadow_advanced_tools.util.zone_plates.fresnel_zone_plate_simulator import FZPAttributes, FZPSimulatorOptions, FresnelZonePlateSimulator

from orangecontrib.shadow.widgets.gui.ow_generic_element import GenericElement

GOOD = 1

COLLIMATED_SOURCE_LIMIT = 1e4 #m

class FresnelZonePlate(GenericElement):

    name = "Fresnel Zone Plate (Beta)"
    description = "Shadow OE: Fresnel Zone Plate"
    icon = "icons/zone_plate.png"
    maintainer = "Luca Rebuffi"
    maintainer_email = "lrebuffi(@at@)anl.gov"
    priority = 23
    category = "Optical Elements"
    keywords = ["data", "file", "load", "read"]

    inputs = [("Input Beam", ShadowBeam, "setBeam")]

    outputs = [{"name":"Beam",
                "type":ShadowBeam,
                "doc":"Shadow Beam",
                "id":"beam"},
               {"name":"Trigger",
                "type": TriggerIn,
                "doc":"Feedback signal to start a new beam simulation",
                "id":"Trigger"}]

    input_beam = None
    output_beam = None

    NONE_SPECIFIED = "NONE SPECIFIED"

    ONE_ROW_HEIGHT = 65
    TWO_ROW_HEIGHT = 110
    THREE_ROW_HEIGHT = 170

    INNER_BOX_WIDTH_L3=322
    INNER_BOX_WIDTH_L2=335
    INNER_BOX_WIDTH_L1=358
    INNER_BOX_WIDTH_L0=375

    source_plane_distance = Setting(10.0)
    image_plane_distance = Setting(20.0)

    height = Setting(20.0) # um
    diameter = Setting(50.0) # um
    b_min = Setting(50.0) # nm
    zone_plate_material = Setting('Au')
    template_material = Setting('SiO2')

    zone_plate_type = Setting(0)
    width_coating = Setting(20) # nm
    height1_factor = Setting(0.33)
    height2_factor = Setting(0.67)

    with_central_stop = Setting(0)
    cs_diameter = Setting(10.0) # um

    with_order_sorting_aperture = Setting(0)

    osa_position = Setting(10.0) # user units
    osa_diameter =  Setting(30.0) # um

    source_distance_flag = Setting(0)
    source_distance = Setting(0.0)

    image_distance_flag = Setting(1)

    with_multi_slicing = Setting(0)
    n_slices = Setting(100)

    n_points = Setting(5000)
    last_index = Setting(100)

    ##################################################

    avg_energy = 0.0
    number_of_zones = 0
    focal_distance = 0.0
    efficiency = 0.0

    ##################################################

    mirror_movement = Setting(0)

    mm_mirror_offset_x = Setting(0.0)
    mm_mirror_rotation_x = Setting(0.0)
    mm_mirror_offset_y = Setting(0.0)
    mm_mirror_rotation_y = Setting(0.0)
    mm_mirror_offset_z = Setting(0.0)
    mm_mirror_rotation_z = Setting(0.0)

    #####

    source_movement = Setting(0)
    sm_angle_of_incidence = Setting(0.0)
    sm_distance_from_mirror = Setting(0.0)
    sm_z_rotation = Setting(0.0)
    sm_offset_x_mirr_ref_frame = Setting(0.0)
    sm_offset_y_mirr_ref_frame = Setting(0.0)
    sm_offset_z_mirr_ref_frame = Setting(0.0)
    sm_offset_x_source_ref_frame = Setting(0.0)
    sm_offset_y_source_ref_frame = Setting(0.0)
    sm_offset_z_source_ref_frame = Setting(0.0)
    sm_rotation_around_x = Setting(0.0)
    sm_rotation_around_y = Setting(0.0)
    sm_rotation_around_z = Setting(0.0)

    #####

    file_to_write_out = Setting(3) # Mirror: users found difficoult to activate the "Footprint" option.
    write_out_inc_ref_angles = Setting(0)

    def __init__(self):
        super(FresnelZonePlate, self).__init__()

        self.runaction = widget.OWAction("Run Shadow/Trace", self)
        self.runaction.triggered.connect(self.traceOpticalElement)
        self.addAction(self.runaction)

        button_box = oasysgui.widgetBox(self.controlArea, "", addSpace=False, orientation="horizontal")

        button = gui.button(button_box, self, "Run Shadow/Trace", callback=self.traceOpticalElement)
        font = QFont(button.font())
        font.setBold(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Blue'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)

        button = gui.button(button_box, self, "Reset Fields", callback=self.callResetSettings)
        font = QFont(button.font())
        font.setItalic(True)
        button.setFont(font)
        palette = QPalette(button.palette()) # make a copy of the palette
        palette.setColor(QPalette.ButtonText, QColor('Dark Red'))
        button.setPalette(palette) # assign new palette
        button.setFixedHeight(45)
        button.setFixedWidth(150)

        gui.separator(self.controlArea)

        tabs_setting = oasysgui.tabWidget(self.controlArea)
        tabs_setting.setFixedHeight(self.TABS_AREA_HEIGHT)
        tabs_setting.setFixedWidth(self.CONTROL_AREA_WIDTH-5)

        tab_pos = oasysgui.createTabPage(tabs_setting, "Position")

        upper_box = oasysgui.widgetBox(tab_pos, "Optical Element Orientation", addSpace=True, orientation="vertical")

        self.le_source_plane_distance = oasysgui.lineEdit(upper_box, self, "source_plane_distance", "Source Plane Distance", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_image_plane_distance  = oasysgui.lineEdit(upper_box, self, "image_plane_distance", "Image Plane Distance", labelWidth=260, valueType=float, orientation="horizontal")

        tab_bas = oasysgui.createTabPage(tabs_setting, "Basic Setting")
        tab_adv = oasysgui.createTabPage(tabs_setting, "Advanced Setting")

        ##########################################
        ##########################################
        # BASIC SETTINGS
        ##########################################
        ##########################################

        tabs_basic_setting = oasysgui.tabWidget(tab_bas)
        tabs_setting.setFixedHeight(self.TABS_AREA_HEIGHT-5)
        tabs_setting.setFixedWidth(self.CONTROL_AREA_WIDTH-5)

        tab_zone_plate_1 = oasysgui.createTabPage(tabs_basic_setting, "Input Parameters")
        tab_zone_plate_2 = oasysgui.createTabPage(tabs_basic_setting, "Propagation Parameters")
        tab_zone_plate_3 = oasysgui.createTabPage(tabs_basic_setting, "Output Parameters")

        zp_box = oasysgui.widgetBox(tab_zone_plate_1, "F.Z.P. Parameters", addSpace=False, orientation="vertical", height=475)

        oasysgui.lineEdit(zp_box, self, "b_min",  "Outermost Zone Width/Period [nm]", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(zp_box, self, "diameter", "F.Z.P. Diameter [" + u"\u03BC" + "m]", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(zp_box, self, "height",  "F.Z.P. Height [" + u"\u03BC" + "m]", labelWidth=260, valueType=float, orientation="horizontal")

        gui.comboBox(zp_box, self, "zone_plate_type", label="Type of F.Z.P.", labelWidth=350,
                     items=["Ordinary", "Zone-Doubled", "Zone-Filled", "Two-Level"],
                     callback=self.set_FZPType, sendSelectedValue=False, orientation="horizontal")

        oasysgui.lineEdit(zp_box, self, "zone_plate_material",  "F.Z.P. Material", labelWidth=160, valueType=str, orientation="horizontal")

        self.ord_box = oasysgui.widgetBox(zp_box, "", addSpace=False, orientation="vertical", height=60)

        self.zd_box = oasysgui.widgetBox(zp_box, "", addSpace=False, orientation="vertical", height=60)
        oasysgui.lineEdit(self.zd_box, self, "template_material",  "Template Material", labelWidth=160, valueType=str, orientation="horizontal")
        oasysgui.lineEdit(self.zd_box, self, "width_coating",  "Coating Width [nm]", labelWidth=260, valueType=float, orientation="horizontal")

        self.zf_box = oasysgui.widgetBox(zp_box, "", addSpace=False, orientation="vertical", height=60)
        oasysgui.lineEdit(self.zf_box, self, "template_material", "Template Material", labelWidth=160, valueType=str, orientation="horizontal")
        oasysgui.lineEdit(self.zf_box, self, "width_coating", "Coating Width [nm]", labelWidth=260, valueType=float, orientation="horizontal")

        self.tl_box = oasysgui.widgetBox(zp_box, "", addSpace=False, orientation="vertical", height=60)
        oasysgui.lineEdit(self.tl_box, self, "height1_factor",  "Height Factor 1", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.tl_box, self, "height2_factor",  "Height Factor 2", labelWidth=260, valueType=float, orientation="horizontal")

        self.set_FZPType()

        gui.comboBox(zp_box, self, "with_central_stop", label="With Central Stop", labelWidth=350,
                     items=["No", "Yes"],
                     callback=self.set_WithCentralStop, sendSelectedValue=False, orientation="horizontal")

        self.cs_box_1 = oasysgui.widgetBox(zp_box, "", addSpace=False, orientation="vertical", height=30)
        self.cs_box_2 = oasysgui.widgetBox(zp_box, "", addSpace=False, orientation="vertical", height=30)

        oasysgui.lineEdit(self.cs_box_1, self, "cs_diameter", "C.S. Diameter [" + u"\u03BC" + "m]", labelWidth=260, valueType=float, orientation="horizontal")

        self.set_WithCentralStop()

        gui.comboBox(zp_box, self, "with_order_sorting_aperture", label="With Order Sorting Aperture", labelWidth=350,
                     items=["No", "Yes"],
                     callback=self.set_WithOrderSortingAperture, sendSelectedValue=False, orientation="horizontal")

        self.osa_box_1 = oasysgui.widgetBox(zp_box, "", addSpace=False, orientation="vertical", height=60)
        self.osa_box_2 = oasysgui.widgetBox(zp_box, "", addSpace=False, orientation="vertical", height=60)

        self.le_osa_position = oasysgui.lineEdit(self.osa_box_1, self, "osa_position", "O.S.A. position ", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.osa_box_1, self, "osa_diameter", "O.S.A. Diameter [" + u"\u03BC" + "m]", labelWidth=260, valueType=float, orientation="horizontal")

        self.set_WithOrderSortingAperture()

        gui.comboBox(zp_box, self, "source_distance_flag", label="Source Distance", labelWidth=350,
                     items=["Same as Source Plane", "Different"],
                     callback=self.set_SourceDistanceFlag, sendSelectedValue=False, orientation="horizontal")

        self.zp_box_1 = oasysgui.widgetBox(zp_box, "", addSpace=False, orientation="vertical", height=30)
        self.zp_box_2 = oasysgui.widgetBox(zp_box, "", addSpace=False, orientation="vertical", height=30)



        self.le_source_distance = oasysgui.lineEdit(self.zp_box_1, self, "source_distance", "Source Distance", labelWidth=260, valueType=float, orientation="horizontal")

        self.set_SourceDistanceFlag()

        gui.comboBox(zp_box, self, "image_distance_flag", label="Image Distance", labelWidth=350,
                     items=["Image Plane Distance", "Z.P. Focal Distance"],
                     callback=self.set_ImageDistanceFlag, sendSelectedValue=False, orientation="horizontal")

        self.set_ImageDistanceFlag()

        prop_box = oasysgui.widgetBox(tab_zone_plate_2, "Propagation Parameters", addSpace=False, orientation="vertical", height=270)

        gui.comboBox(prop_box, self, "with_multi_slicing", label="With Multi-Slicing", labelWidth=350,
                     items=["No", "Yes"],
                     callback=self.set_WithMultislicing, sendSelectedValue=False, orientation="horizontal")

        self.ms_box_1 = oasysgui.widgetBox(prop_box, "", addSpace=False, orientation="vertical", height=30)
        self.ms_box_2 = oasysgui.widgetBox(prop_box, "", addSpace=False, orientation="vertical", height=30)

        oasysgui.lineEdit(self.ms_box_1, self, "n_slices", "Nr. Slices", labelWidth=260, valueType=int, orientation="horizontal")

        self.set_WithMultislicing()

        oasysgui.lineEdit(prop_box, self, "n_points", "Nr. Sampling Points", labelWidth=260, valueType=int, orientation="horizontal")
        oasysgui.lineEdit(prop_box, self, "last_index", "Last Index of Focal Image", labelWidth=260, valueType=int, orientation="horizontal")

        zp_out_box = oasysgui.widgetBox(tab_zone_plate_3, "Output Parameters", addSpace=False, orientation="vertical", height=270)

        self.le_avg_wavelength = oasysgui.lineEdit(zp_out_box, self, "avg_energy", "Average Energy [eV]", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_avg_wavelength.setReadOnly(True)
        font = QFont(self.le_avg_wavelength.font())
        font.setBold(True)
        self.le_avg_wavelength.setFont(font)
        palette = QPalette(self.le_avg_wavelength.palette()) # make a copy of the palette
        palette.setColor(QPalette.Text, QColor('dark blue'))
        palette.setColor(QPalette.Base, QColor(243, 240, 160))
        self.le_avg_wavelength.setPalette(palette)

        self.le_number_of_zones = oasysgui.lineEdit(zp_out_box, self, "number_of_zones", "Number of Zones", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_number_of_zones.setReadOnly(True)
        font = QFont(self.le_number_of_zones.font())
        font.setBold(True)
        self.le_number_of_zones.setFont(font)
        palette = QPalette(self.le_number_of_zones.palette()) # make a copy of the palette
        palette.setColor(QPalette.Text, QColor('dark blue'))
        palette.setColor(QPalette.Base, QColor(243, 240, 160))
        self.le_number_of_zones.setPalette(palette)

        self.le_focal_distance = oasysgui.lineEdit(zp_out_box, self, "focal_distance", "Focal Distance", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_focal_distance.setReadOnly(True)
        font = QFont(self.le_focal_distance.font())
        font.setBold(True)
        self.le_focal_distance.setFont(font)
        palette = QPalette(self.le_focal_distance.palette()) # make a copy of the palette
        palette.setColor(QPalette.Text, QColor('dark blue'))
        palette.setColor(QPalette.Base, QColor(243, 240, 160))
        self.le_focal_distance.setPalette(palette)

        self.le_efficiency = oasysgui.lineEdit(zp_out_box, self, "efficiency", "Efficiency % (Avg. Wavelength)", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_efficiency.setReadOnly(True)
        font = QFont(self.le_efficiency.font())
        font.setBold(True)
        self.le_efficiency.setFont(font)
        palette = QPalette(self.le_efficiency.palette()) # make a copy of the palette
        palette.setColor(QPalette.Text, QColor('dark blue'))
        palette.setColor(QPalette.Base, QColor(243, 240, 160))
        self.le_efficiency.setPalette(palette)


        ##########################################
        ##########################################
        # ADVANCED SETTINGS
        ##########################################
        ##########################################

        tabs_advanced_setting = oasysgui.tabWidget(tab_adv)

        tab_adv_mir_mov = oasysgui.createTabPage(tabs_advanced_setting, "O.E. Movement")
        tab_adv_sou_mov = oasysgui.createTabPage(tabs_advanced_setting, "Source Movement")
        tab_adv_misc = oasysgui.createTabPage(tabs_advanced_setting, "Output Files")

        ##########################################
        #
        # TAB 2.2 - Mirror Movement
        #
        ##########################################

        mir_mov_box = oasysgui.widgetBox(tab_adv_mir_mov, "O.E. Movement Parameters", addSpace=False, orientation="vertical", height=230)

        gui.comboBox(mir_mov_box, self, "mirror_movement", label="O.E. Movement", labelWidth=350,
                     items=["No", "Yes"],
                     callback=self.set_MirrorMovement, sendSelectedValue=False, orientation="horizontal")

        gui.separator(mir_mov_box, height=10)

        self.mir_mov_box_1 = oasysgui.widgetBox(mir_mov_box, "", addSpace=False, orientation="vertical")

        self.le_mm_mirror_offset_x = oasysgui.lineEdit(self.mir_mov_box_1, self, "mm_mirror_offset_x", "O.E. Offset X", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.mir_mov_box_1, self, "mm_mirror_rotation_x", "O.E. Rotation X [CCW, deg]", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_mm_mirror_offset_y = oasysgui.lineEdit(self.mir_mov_box_1, self, "mm_mirror_offset_y", "O.E. Offset Y", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.mir_mov_box_1, self, "mm_mirror_rotation_y", "O.E. Rotation Y [CCW, deg]", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_mm_mirror_offset_z = oasysgui.lineEdit(self.mir_mov_box_1, self, "mm_mirror_offset_z", "O.E. Offset Z", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.mir_mov_box_1, self, "mm_mirror_rotation_z", "O.E. Rotation Z [CCW, deg]", labelWidth=260, valueType=float, orientation="horizontal")

        self.set_MirrorMovement()

       ##########################################
        #
        # TAB 2.3 - Source Movement
        #
        ##########################################

        sou_mov_box = oasysgui.widgetBox(tab_adv_sou_mov, "Source Movement Parameters", addSpace=False, orientation="vertical", height=400)

        gui.comboBox(sou_mov_box, self, "source_movement", label="Source Movement", labelWidth=350,
                     items=["No", "Yes"],
                     callback=self.set_SourceMovement, sendSelectedValue=False, orientation="horizontal")

        gui.separator(sou_mov_box, height=10)

        self.sou_mov_box_1 = oasysgui.widgetBox(sou_mov_box, "", addSpace=False, orientation="vertical")

        oasysgui.lineEdit(self.sou_mov_box_1, self, "sm_angle_of_incidence", "Angle of Incidence [deg]", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_sm_distance_from_mirror = oasysgui.lineEdit(self.sou_mov_box_1, self, "sm_distance_from_mirror", "Distance from O.E.", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.sou_mov_box_1, self, "sm_z_rotation", "Z-rotation [deg]", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_sm_offset_x_mirr_ref_frame = oasysgui.lineEdit(self.sou_mov_box_1, self, "sm_offset_x_mirr_ref_frame", "--", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_sm_offset_y_mirr_ref_frame = oasysgui.lineEdit(self.sou_mov_box_1, self, "sm_offset_y_mirr_ref_frame", "--", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_sm_offset_z_mirr_ref_frame = oasysgui.lineEdit(self.sou_mov_box_1, self, "sm_offset_z_mirr_ref_frame", "--", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_sm_offset_x_source_ref_frame = oasysgui.lineEdit(self.sou_mov_box_1, self, "sm_offset_x_source_ref_frame", "--", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_sm_offset_y_source_ref_frame = oasysgui.lineEdit(self.sou_mov_box_1, self, "sm_offset_y_source_ref_frame", "--", labelWidth=260, valueType=float, orientation="horizontal")
        self.le_sm_offset_z_source_ref_frame = oasysgui.lineEdit(self.sou_mov_box_1, self, "sm_offset_z_source_ref_frame", "--", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.sou_mov_box_1, self, "sm_rotation_around_x", "rotation [CCW, deg] around X", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.sou_mov_box_1, self, "sm_rotation_around_y", "rotation [CCW, deg] around Y", labelWidth=260, valueType=float, orientation="horizontal")
        oasysgui.lineEdit(self.sou_mov_box_1, self, "sm_rotation_around_z", "rotation [CCW, deg] around Z", labelWidth=260, valueType=float, orientation="horizontal")

        self.set_SourceMovement()

        ##########################################
        #
        # TAB 2.4 - Other
        #
        ##########################################

        adv_other_box = oasysgui.widgetBox(tab_adv_misc, "Optional file output", addSpace=False, orientation="vertical")

        gui.comboBox(adv_other_box, self, "file_to_write_out", label="Files to write out", labelWidth=150,
                     items=["All", "Mirror", "Image", "None", "Debug (All + start.xx/end.xx)"],
                     sendSelectedValue=False, orientation="horizontal")

        gui.comboBox(adv_other_box, self, "write_out_inc_ref_angles", label="Write out Incident/Reflected angles [angle.xx]", labelWidth=300,
                     items=["No", "Yes"],
                     sendSelectedValue=False, orientation="horizontal")


        gui.rubber(self.controlArea)
        gui.rubber(self.mainArea)

        propagation_plot_tab = oasysgui.widgetBox(self.main_tabs, addToLayout=0, margin=4)

        self.main_tabs.insertTab(1, propagation_plot_tab, "TEMP")
        self.main_tabs.setTabText(0, "Shadow Plot")
        self.main_tabs.setTabText(1, "F.Z.P. Simulator Plot")

        self.prop_tabs = oasysgui.tabWidget(propagation_plot_tab)
        self.prop_tab = [oasysgui.createTabPage(self.prop_tabs, "Radial Intensity"),
                         oasysgui.createTabPage(self.prop_tabs, "Generated 2D distribution")]
        self.prop_plot_canvas = [None, None]

    def isFootprintEnabled(self):
        return False

    def enableFootprint(self, enabled=False):
        pass

    def traceOpticalElement(self):
        try:
            self.setStatusMessage("")
            self.progressBarInit()

            if ShadowCongruence.checkEmptyBeam(self.input_beam):
                if ShadowCongruence.checkGoodBeam(self.input_beam):
                    self.checkFields()

                    sys.stdout = EmittingStream(textWritten=self.writeStdOut)

                    if self.trace_shadow:
                        grabber = TTYGrabber()
                        grabber.start()

                    ###########################################
                    # TODO: TO BE ADDED JUST IN CASE OF BROKEN
                    #       ENVIRONMENT: MUST BE FOUND A PROPER WAY
                    #       TO TEST SHADOW
                    self.fixWeirdShadowBug()
                    ###########################################

                    self.progressBarSet(10)

                    if self.source_distance_flag == 0:
                        self.source_distance = self.source_plane_distance

                    options = FZPSimulatorOptions(with_central_stop=self.with_central_stop==1,
                                                  cs_diameter=numpy.round(self.cs_diameter*1e-6, 7),
                                                  with_order_sorting_aperture=self.with_order_sorting_aperture==1,
                                                  osa_position=self.osa_position*self.workspace_units_to_m,
                                                  osa_diameter=numpy.round(self.osa_diameter*1e-6, 7),
                                                  zone_plate_type=self.zone_plate_type,
                                                  width_coating=numpy.round(self.width_coating*1e-9, 10),
                                                  height1_factor=self.height1_factor,
                                                  height2_factor=self.height2_factor,
                                                  with_range=False,
                                                  with_multi_slicing=self.with_multi_slicing==1,
                                                  n_slices=self.n_slices,
                                                  with_complex_amplitude=False)
                    attributes = FZPAttributes(height=numpy.round(self.height*1e-6, 7),
                                               diameter=numpy.round(self.diameter*1e-6, 7),
                                               b_min=numpy.round(self.b_min*1e-9, 10),
                                               zone_plate_material=self.zone_plate_material,
                                               template_material=self.template_material)

                    zone_plate_beam = self.get_zone_plate_beam(attributes)

                    self.progressBarSet(30)

                    self.avg_energy = numpy.round(ShadowPhysics.getEnergyFromShadowK(numpy.average(zone_plate_beam._beam.rays[numpy.where(zone_plate_beam._beam.rays[:, 9] == GOOD), 10])), 2)

                    fzp_simulator = FresnelZonePlateSimulator(attributes=attributes, options=options)
                    fzp_simulator.initialize(energy_in_KeV=self.avg_energy/1000, n_points=self.n_points)

                    self.number_of_zones = fzp_simulator.n_zones
                    self.focal_distance = numpy.round(fzp_simulator.focal_distance/self.workspace_units_to_m, 5)

                    beam_out = self.get_output_beam(zone_plate_beam, fzp_simulator)

                    self.progressBarSet(80)

                    if self.trace_shadow:
                        grabber.stop()

                        for row in grabber.ttyData:
                           self.writeStdOut(row)

                    self.setStatusMessage("Plotting Results")

                    self.plot_results(beam_out)

                    self.setStatusMessage("")

                    beam_out.setScanningData(self.input_beam.scanned_variable_data)

                    self.send("Beam", beam_out)
                    self.send("Trigger", TriggerIn(new_object=True))
                else:
                    raise Exception("Input Beam with no good rays")
            else:
                raise Exception("Empty Input Beam")

        except Exception as exception:
            QMessageBox.critical(self, "Error", str(exception), QMessageBox.Ok)

            if self.IS_DEVELOP: raise exception

        self.progressBarFinished()

    def setBeam(self, beam):
        if ShadowCongruence.checkEmptyBeam(beam):
            self.input_beam = beam

            if self.is_automatic_run:
                self.traceOpticalElement()

    def checkFields(self):
        self.source_plane_distance = congruence.checkNumber(self.source_plane_distance, "Source plane distance")
        self.image_plane_distance = congruence.checkNumber(self.image_plane_distance, "Image plane distance")


    def after_change_workspace_units(self):
        label = self.le_source_plane_distance.parent().layout().itemAt(0).widget()
        label.setText(label.text() + " [" + self.workspace_units_label + "]")
        label = self.le_image_plane_distance.parent().layout().itemAt(0).widget()
        label.setText(label.text() + " [" + self.workspace_units_label + "]")

        label = self.le_osa_position.parent().layout().itemAt(0).widget()
        label.setText(label.text() + " [" + self.workspace_units_label + "]")
        label = self.le_source_distance.parent().layout().itemAt(0).widget()
        label.setText(label.text() + " [" + self.workspace_units_label + "]")
        label = self.le_focal_distance.parent().layout().itemAt(0).widget()
        label.setText(label.text() + " [" + self.workspace_units_label + "]")

        # ADVANCED SETTINGS
        # MIRROR MOVEMENTS
        label = self.le_mm_mirror_offset_x.parent().layout().itemAt(0).widget()
        label.setText(label.text() + " [" + self.workspace_units_label + "]")
        label = self.le_mm_mirror_offset_y.parent().layout().itemAt(0).widget()
        label.setText(label.text() + " [" + self.workspace_units_label + "]")
        label = self.le_mm_mirror_offset_z.parent().layout().itemAt(0).widget()
        label.setText(label.text() + " [" + self.workspace_units_label + "]")
        # SOURCE MOVEMENTS
        label = self.le_sm_distance_from_mirror.parent().layout().itemAt(0).widget()
        label.setText(label.text() + " [" + self.workspace_units_label + "]")
        label = self.le_sm_offset_x_mirr_ref_frame.parent().layout().itemAt(0).widget()
        label.setText("offset X [" + self.workspace_units_label + "] in O.E. reference frame")
        label = self.le_sm_offset_y_mirr_ref_frame.parent().layout().itemAt(0).widget()
        label.setText("offset Y [" + self.workspace_units_label + "] in O.E. reference frame")
        label = self.le_sm_offset_z_mirr_ref_frame.parent().layout().itemAt(0).widget()
        label.setText("offset Z [" + self.workspace_units_label + "] in O.E. reference frame")
        label = self.le_sm_offset_x_source_ref_frame.parent().layout().itemAt(0).widget()
        label.setText("offset X [" + self.workspace_units_label + "] in SOURCE reference frame")
        label = self.le_sm_offset_y_source_ref_frame.parent().layout().itemAt(0).widget()
        label.setText("offset Y [" + self.workspace_units_label + "] in SOURCE reference frame")
        label = self.le_sm_offset_z_source_ref_frame.parent().layout().itemAt(0).widget()
        label.setText("offset Z [" + self.workspace_units_label + "] in SOURCE reference frame")

    def callResetSettings(self):
        super().callResetSettings()
        self.setupUI()

    def set_SourceMovement(self):
        self.sou_mov_box_1.setVisible(self.source_movement == 1)

    def set_MirrorMovement(self):
        self.mir_mov_box_1.setVisible(self.mirror_movement == 1)

    def set_FZPType(self):
        self.ord_box.setVisible(self.zone_plate_type == 0)
        self.zd_box.setVisible(self.zone_plate_type == 1)
        self.zf_box.setVisible(self.zone_plate_type == 2)
        self.tl_box.setVisible(self.zone_plate_type == 3)

    def set_SourceDistanceFlag(self):
        self.zp_box_1.setVisible(self.source_distance_flag == 1)
        self.zp_box_2.setVisible(self.source_distance_flag == 0)

    def set_WithCentralStop(self):
        self.cs_box_1.setVisible(self.with_central_stop == 1)
        self.cs_box_2.setVisible(self.with_central_stop == 0)

    def set_WithOrderSortingAperture(self):
        self.osa_box_1.setVisible(self.with_order_sorting_aperture == 1)
        self.osa_box_2.setVisible(self.with_order_sorting_aperture == 0)

    def set_WithMultislicing(self):
        self.ms_box_1.setVisible(self.with_multi_slicing == 1)
        self.ms_box_2.setVisible(self.with_multi_slicing == 0)

    def set_ImageDistanceFlag(self):
        self.le_image_plane_distance.setEnabled(self.image_distance_flag==0)

    ######################################################################
    # ZONE PLATE CALCULATION
    ######################################################################

    def get_zone_plate_beam(self, attributes):       # WS Units

        empty_element = ShadowOpticalElement.create_empty_oe()

        empty_element._oe.DUMMY        = self.workspace_units_to_cm
        empty_element._oe.T_SOURCE     = self.source_plane_distance
        empty_element._oe.T_IMAGE      = 0.0
        empty_element._oe.T_INCIDENCE  = 0.0
        empty_element._oe.T_REFLECTION = 180.0
        empty_element._oe.ALPHA        = 0.0

        empty_element._oe.FWRITE = 3
        empty_element._oe.F_ANGLE = 0

        n_screen = 1
        i_screen = numpy.array([0, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        i_abs = numpy.zeros(10)
        i_slit = numpy.array([1, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        i_stop = numpy.zeros(10)
        k_slit = numpy.array([1, 0, 0, 0, 0, 0, 0, 0, 0, 0])
        thick = numpy.zeros(10)
        file_abs = numpy.array(['', '', '', '', '', '', '', '', '', ''])
        rx_slit = numpy.zeros(10)
        rz_slit = numpy.zeros(10)
        sl_dis = numpy.zeros(10)
        file_scr_ext = numpy.array(['', '', '', '', '', '', '', '', '', ''])
        cx_slit = numpy.zeros(10)
        cz_slit = numpy.zeros(10)

        sl_dis[0] = 0.0
        rx_slit[0] = attributes.diameter/self.workspace_units_to_m
        rz_slit[0] = attributes.diameter/self.workspace_units_to_m
        cx_slit[0] = 0.0
        cz_slit[0] = 0.0

        empty_element._oe.set_screens(n_screen,
                                      i_screen,
                                      i_abs,
                                      sl_dis,
                                      i_slit,
                                      i_stop,
                                      k_slit,
                                      thick,
                                      file_abs,
                                      rx_slit,
                                      rz_slit,
                                      cx_slit,
                                      cz_slit,
                                      file_scr_ext)

        output_beam = ShadowBeam.traceFromOE(self.input_beam, empty_element, history=True)

        return output_beam

    def get_output_beam(self, zone_plate_beam, fzp_simulator):
        # ----------------------------------------------------------------------------------------
        # raytracing of an ideal lens to focal position
        ideal_lens = ShadowOpticalElement.create_ideal_lens()

        focal_distance = fzp_simulator.focal_distance/self.workspace_units_to_m
        focal_xz = 1 / ((1 / self.source_distance) + (1 / focal_distance))

        ideal_lens._oe.focal_x = focal_xz
        ideal_lens._oe.focal_z = focal_xz

        ideal_lens._oe.user_units_to_cm = self.workspace_units_to_cm
        ideal_lens._oe.T_SOURCE         = 0.0
        ideal_lens._oe.T_IMAGE          = focal_distance

        output_beam = ShadowBeam.traceIdealLensOE(zone_plate_beam, ideal_lens, history=True)

        # ----------------------------------------------------------------------------------------

        intensity, _, efficiency = fzp_simulator.simulate()

        self.efficiency = numpy.round(efficiency*100, 3)

        #----------------------------------------------------------------------------------------
        # from Hybrid: the ideal focusing is corrected by using the image at focus as a divergence correction distribution
        intensity_index = self.n_slices if self.with_multi_slicing else 1

        X, Y, dif_xpzp = fzp_simulator.create_2D_profile(intensity[intensity_index, :], last_index=self.last_index)
        xp = X[0, :]/fzp_simulator.focal_distance
        zp = Y[:, 0]/fzp_simulator.focal_distance

        go = numpy.where(output_beam._beam.rays[:, 9] == GOOD)

        dx_ray = numpy.arctan(output_beam._beam.rays[go, 3] / output_beam._beam.rays[go, 4])  # calculate divergence from direction cosines from SHADOW file  dx = atan(v_x/v_y)
        dz_ray = numpy.arctan(output_beam._beam.rays[go, 5] / output_beam._beam.rays[go, 4])  # calculate divergence from direction cosines from SHADOW file  dz = atan(v_z/v_y)

        s2d = Sampler2D(dif_xpzp, xp, zp)

        pos_dif_x, pos_dif_z = s2d.get_n_sampled_points(dx_ray.shape[1])

        #correction to the position with the divergence kick from the waveoptics calculation
        xx_image = output_beam._beam.rays[go, 0] + focal_distance * pos_dif_x # ray tracing to the image plane
        zz_image = output_beam._beam.rays[go, 2] + focal_distance * pos_dif_z # ray tracing to the image plane

        output_beam._oe_number = self.input_beam._oe_number + 1

        # new divergence distribution: convolution
        dx_conv = numpy.arctan(pos_dif_x) + dx_ray  # add the ray divergence kicks
        dz_conv = numpy.arctan(pos_dif_z) + dz_ray  # add the ray divergence kicks

        angle_num = numpy.sqrt(1 + (numpy.tan(dz_conv)) ** 2 + (numpy.tan(dx_conv)) ** 2)

        output_beam._beam.rays[go, 0] = copy.deepcopy(xx_image)
        output_beam._beam.rays[go, 2] = copy.deepcopy(zz_image)
        output_beam._beam.rays[go, 3] = numpy.tan(dx_conv) / angle_num
        output_beam._beam.rays[go, 4] = 1 / angle_num
        output_beam._beam.rays[go, 5] = numpy.tan(dz_conv) / angle_num
        #----------------------------------------------------------------------------------------

        output_beam._beam.rays[go, 6] *= numpy.sqrt(efficiency)
        output_beam._beam.rays[go, 7] *= numpy.sqrt(efficiency)
        output_beam._beam.rays[go, 8] *= numpy.sqrt(efficiency)
        output_beam._beam.rays[go, 15] *= numpy.sqrt(efficiency)
        output_beam._beam.rays[go, 16] *= numpy.sqrt(efficiency)
        output_beam._beam.rays[go, 17] *= numpy.sqrt(efficiency)

        if self.image_distance_flag==0: output_beam._beam.retrace(self.image_plane_distance - focal_distance)

        self. plot_propagation_results(fzp_simulator, intensity[intensity_index, :])

        return output_beam

    def plot_propagation_results(self, fzp_simulator, profile_1D):
        if self.prop_plot_canvas[0] is None:
            self.prop_plot_canvas[0] = fzp_simulator.plot_1D(None, profile_1D, self.last_index, replace=True)
            self.prop_tab[0].layout().addWidget(self.prop_plot_canvas[0])
        else:
            fzp_simulator.plot_1D(self.prop_plot_canvas[0], profile_1D, self.last_index, replace=True)

        if self.prop_plot_canvas[1] is None:
            self.prop_plot_canvas[1] = fzp_simulator.plot_2D(None, profile_1D, self.last_index)
            self.prop_tab[1].layout().addWidget(self.prop_plot_canvas[1])
        else:
            fzp_simulator.plot_2D(self.prop_plot_canvas[1], profile_1D, self.last_index)




# #########################################################################
# Copyright (c) 2020, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2020. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# #########################################################################
# %% Fresnel zone plate simulation code
# %--------------------------------------------------------------------------
# % by Joan Vila-Comamala from original IDL version of Ana Diaz (February, 2009)
# % June, 2010
# %
# % code modified by Michael Wojcik Oct, 2013
# %
# % It simulates wavefront after FZP and propagates to the focal plane
# % plots wavefield all through the propagation from the FZP to the focus.
# %
# % 2D Extension Code --> Wavefront propagation in made using the Hankel
# % transform for a circularly symmetric function a long the radial
# % coordinate. Hankel transform routine is Hankel_Transform_MGS.
# %
# % Keep your eyes open, this code has not been throughly debugged or tested!
# %
# %
# %% ------------------------------------------------------------------------

import numpy

from matplotlib import cm, rcParams
from matplotlib.backends.backend_qt5agg import FigureCanvasQTAgg
from matplotlib.figure import Figure

try:
    from mpl_toolkits.mplot3d import Axes3D  # to load plot 3D
except:
    pass

from scipy import interpolate

from oasys.widgets import gui as oasysgui
from silx.gui.plot import Plot2D

from orangecontrib.shadow_advanced_tools.util.zone_plates import bessel_zeros
from orangecontrib.shadow_advanced_tools.util.zone_plates.hankel_transform import hankel_transform
from orangecontrib.shadow_advanced_tools.util.zone_plates.refractive_index import get_delta_beta


# % ------------------------------------------
#
# cs_diameter = beamstop diameter [m]
# osa_position = distance FZP-OSA [m]
# osa_diameter = OSA diameter [m]
#
# zone_plate_type = 0  # equal to 0 --> Ordinary FZP
#                      # equal to 1 --> Zone-Doubled FZP
#                      # equal to 2 --> Zone-Filled FZP
#                      # equal to 3 --> Two-Level FZP
#                      # equal to 4 --> Three-Level FZP           - not implemented
#                      # equal to 5 --> ALD Multideposition FZP 1 - not implemented
#                      # equal to 6 --> ALD Multideposition FZP 2 - not implemented
#                      # equal to 7 --> Zone-Edge Slanted FZP     - not implemented
#
# width_coating  = Width of the coated material for a zone-filled FZP
#
# height1_factor = multiply height < 1 for Two-level profile
# height2_factor = multiply height < 1 for Two-level profile
#
# with_range =  False --> plot to focal length
#               True  --> plot in a given range
#
# range_i  = initial position of the range
# range_f  = final position of the range
# n_z      = number of positions
#
# n_slices           = number of slices
# with_complex_amplitude = False
#
# % ------------------------------------------
class FZPSimulatorOptions():

    def __init__(self,
                 with_central_stop=True,
                 cs_diameter=10e-6,
                 with_order_sorting_aperture=False,
                 osa_position=0.01,
                 osa_diameter=30e-6,
                 zone_plate_type=0,
                 width_coating=20e-9,
                 height1_factor=(1 / 3),
                 height2_factor=(2 / 3),
                 with_range=False,
                 range_i=-2e-6,
                 range_f=2e-6,
                 n_z=3,
                 with_multi_slicing=False,
                 n_slices=100,
                 with_complex_amplitude=False
                 ):
        self.with_central_stop = with_central_stop
        self.cs_diameter = cs_diameter
        self.with_order_sorting_aperture = with_order_sorting_aperture
        self.osa_position = osa_position
        self.osa_diameter = osa_diameter
        self.zone_plate_type = zone_plate_type
        self.width_coating = width_coating
        self.height1_factor = height1_factor
        self.height2_factor = height2_factor
        self.with_range = with_range
        self.range_i = range_i
        self.range_f = range_f
        self.n_z = n_z
        self.with_multi_slicing = with_multi_slicing
        self.n_slices = n_slices
        self.with_complex_amplitude = with_complex_amplitude


# % ------------------------------------------
#
# height =  zone thickness or height [m]
# diameter = FZP diameter [m]
# b_min = outermost zone width [m] / outermost period for ZD [m]
#
# % ------------------------------------------
class FZPAttributes():

    def __init__(self,
                 height=20000e-9,
                 diameter=50e-6,
                 b_min=50e-9,
                 zone_plate_material='Au',
                 template_material='SiO2'
                 ):
        self.height = height
        self.diameter = diameter
        self.b_min = b_min
        self.zone_plate_material = zone_plate_material
        self.template_material = template_material


class FresnelZonePlateSimulator(object):

    def __init__(self, options=FZPSimulatorOptions(), attributes=FZPAttributes()):
        self.__options = options
        self.__attributes = attributes

    def initialize(self, energy_in_KeV=8.0, n_points=5000):
        at = self.__attributes
        op = self.__options

        if energy_in_KeV <= 0: raise ValueError("Energy must be > 0")
        if n_points <= 0: raise ValueError("Number of integration points must be > 0")

        if at.height <= 0: raise ValueError("ZP Height must be > 0")
        if at.b_min <= 0: raise ValueError("ZP outermost zone width must be > 0")
        if at.diameter <= 0: raise ValueError("ZP Diameter must be > 0")

        if op.zone_plate_type in [1, 2]:
            if op.width_coating <= 0: raise ValueError("Coating Width must be > 0")
        if op.zone_plate_type == 3:
            if op.height1_factor <= 0: raise ValueError("Height 1 Factor must be > 0")
            if op.height2_factor <= 0: raise ValueError("Height 2 Factor must be > 0")

        self.energy_in_KeV = energy_in_KeV
        self.n_points = n_points

        self.wavelength = 12.398 / energy_in_KeV * 1e-10  # wavelength [m]
        self.k = 2 * numpy.pi / self.wavelength  # wavevector [m-1]

        self.focal_distance = at.diameter * at.b_min / self.wavelength  # focal distance [m]
        self.n_zones = int(numpy.floor(1.0 / 4.0 * (at.diameter / at.b_min)))

        self.max_radius = at.diameter
        self.step = self.max_radius / self.n_points

        self.n_zeros = int(numpy.floor(1.25 * at.diameter / 2 / self.max_radius * n_points))  # Parameter to speed up the Hankel Transform
        # when the function has zeros for N > Nzero

        self.delta_FZP, self.beta_FZP = get_delta_beta(energy_in_KeV, at.zone_plate_material)
        self.delta_template, self.beta_template = get_delta_beta(energy_in_KeV, at.template_material)

        if op.with_multi_slicing:
            if op.n_slices <= 1: raise ValueError("Number of slices position must be > 1")
            self.n_slices = op.n_slices
        else:
            self.n_slices = 1

        if op.with_range:
            if op.range_f <= op.range_i: raise ValueError("Range final position is smaller than initial position")
            if op.n_z < 2: raise ValueError("Number of position must be >= 2")

            self.n_z = op.n_z
        else:
            self.n_z = 1

        if op.with_order_sorting_aperture:
            if op.osa_position <= 0: raise ValueError("OSA position must be > 0")
            if op.osa_position >= self.focal_distance: raise ValueError("OSA position beyond focal distance")
            if op.osa_diameter <= 0: raise ValueError("OSA diameter must be > 0")

    def simulate(self):
        at = self.__attributes
        op = self.__options

        profile, membrane_transmission = self.__build_zone_plate_profile()

        # Loading the position of the zeros, as much position as N+1. The file
        # c.mat contains up to 200000 zeros of the 1st order Bessel function.
        c = bessel_zeros['c'][0, :self.n_points + 1]

        # Definition of the position where the calculated input and transformed
        # functions are evaluated. We define also the maximum frequency in the
        # angular domain.
        q_max = c[self.n_points] / (2 * numpy.pi * self.max_radius) # Maximum frequency
        r = c[:self.n_points] * self.max_radius / c[self.n_points]  # Radius vector
        q = c[:self.n_points] / (2 * numpy.pi * self.max_radius)    # Frequency vector

        # Recalculation of the position where the initial profile is defined.
        profile_h = self.__get_profile_h(profile, r)

        map_int = numpy.zeros((self.n_slices + self.n_z, self.n_points))
        map_complex = numpy.full((self.n_slices + self.n_z, self.n_points), 0j)

        # Calculation of the first angular spectrum
        # --------------------------------------------------------------------------

        print("Initialization, (or Slice #: ", 1, ")" )

        field0 = profile_h * membrane_transmission
        map_int[0, :] = numpy.multiply(numpy.abs(field0), numpy.abs(field0))
        map_complex[0, :] = field0[0: self.n_points]
        four0 = hankel_transform(field0, self.max_radius, c)
        field0 = profile_h

        if op.with_multi_slicing: four0 = self.__propagate_multislicing(map_int, map_complex, field0, four0, q_max, q, c)

        if op.with_range:
            self.__propagate_on_range(map_int, map_complex, four0, q_max, q, c)
        else:
            self.__propagate_to_focus(map_int, map_complex, four0, q_max, q, c)

        efficiency = self.__calculate_efficiency(-1, map_int, profile_h, r, int(numpy.floor(10 * at.b_min / self.step)))

        return map_int, map_complex, efficiency

    def plot_1D(self, plot_canvas, profile_1D, last_index=-1, show=False, replace=True, profile_name="z pos #1", control=False, color='blue'):
        if plot_canvas is None:
            plot_canvas = oasysgui.plotWindow(parent=None,
                                              backend=None,
                                              resetzoom=True,
                                              autoScale=True,
                                              logScale=True,
                                              grid=True,
                                              curveStyle=True,
                                              colormap=False,
                                              aspectRatio=False,
                                              yInverted=False,
                                              copy=True,
                                              save=True,
                                              print_=True,
                                              control=control,
                                              position=True,
                                              roi=False,
                                              mask=False,
                                              fit=True)

            plot_canvas.setDefaultPlotLines(True)
            plot_canvas.setActiveCurveColor(color="#00008B")

        title  = "Radial Intensity Profile"
        xtitle = "Radius [m]"
        ytitle = "Intensity [A.U.]"

        plot_canvas.setGraphTitle(title)
        plot_canvas.setGraphXLabel(xtitle)
        plot_canvas.setGraphYLabel(ytitle)

        rcParams['axes.formatter.useoffset']='False'

        radius = numpy.arange(0, self.max_radius, self.step)

        plot_canvas.addCurve(radius[:last_index], profile_1D[:last_index], profile_name, symbol='', color=color, xlabel=xtitle, ylabel=ytitle, replace=replace) #'+', '^', ','

        plot_canvas.setInteractiveMode('zoom', color='orange')
        plot_canvas.resetZoom()
        plot_canvas.replot()

        plot_canvas.setActiveCurve("Radial Intensity Profile")

        if show: plot_canvas.show()

        return plot_canvas

    def plot_2D(self, plot_canvas, profile_1D, last_index=-1, show=False):
        radius = numpy.arange(0, self.max_radius, self.step)

        X, Y, data2D = FresnelZonePlateSimulator.__create_2D_profile_from_1D(radius[:last_index], profile_1D[:last_index])
        dataX = X[0, :]
        dataY = Y[:, 0]

        origin = (dataX[0], dataY[0])
        scale = (dataX[1] - dataX[0], dataY[1] - dataY[0])

        data_to_plot = data2D.T

        colormap = {"name": "temperature", "normalization": "linear", "autoscale": True, "vmin": 0, "vmax": 0, "colors": 256}

        if plot_canvas is None:
            plot_canvas = Plot2D()

            plot_canvas.resetZoom()
            plot_canvas.setXAxisAutoScale(True)
            plot_canvas.setYAxisAutoScale(True)
            plot_canvas.setGraphGrid(False)
            plot_canvas.setKeepDataAspectRatio(True)
            plot_canvas.yAxisInvertedAction.setVisible(False)

            plot_canvas.setXAxisLogarithmic(False)
            plot_canvas.setYAxisLogarithmic(False)

            plot_canvas.getMaskAction().setVisible(False)
            plot_canvas.getRoiAction().setVisible(False)
            plot_canvas.getColormapAction().setVisible(True)
            plot_canvas.setKeepDataAspectRatio(False)

        plot_canvas.clear()
        plot_canvas.addImage(numpy.array(data_to_plot),
                                                legend="rotated",
                                                scale=scale,
                                                origin=origin,
                                                colormap=colormap,
                                                replace=True)

        plot_canvas.setActiveImage("rotated")
        plot_canvas.setGraphXLabel("X [m]")
        plot_canvas.setGraphYLabel("Y [m]")
        plot_canvas.setGraphTitle("2D Intensity Profile")

        if show: plot_canvas.show()

        return plot_canvas

    def plot_3D(self, figure_canvas, profile_1D, last_index=-1, show=False):
        radius = numpy.arange(0, self.max_radius, self.step)

        X, Y, data2D = FresnelZonePlateSimulator.__create_2D_profile_from_1D(radius[:last_index], profile_1D[:last_index])

        if figure_canvas is None:
            figure = Figure(figsize=(600, 600))
            figure.patch.set_facecolor('white')

            ax = figure.add_subplot(111, projection='3d')
            figure_canvas = FigureCanvasQTAgg(figure)
        else:
            ax = figure_canvas.figure.axes[0]

        ax.clear()

        figure.suptitle("3D Intensity Profile")
        ax.plot_surface(X, Y, data2D, rstride=1, cstride=1, cmap=cm.coolwarm, linewidth=0.5, antialiased=True)

        ax.set_xlabel("X [m]")
        ax.set_ylabel("Y [m]")
        ax.set_zlabel("Intensity [A.U.]")
        ax.mouse_init()

        if show: figure_canvas.show()

        return figure_canvas

    def create_2D_profile(self, profile_1D, last_index=-1):
        radius = numpy.arange(0, self.max_radius, self.step)

        return FresnelZonePlateSimulator.__create_2D_profile_from_1D(radius[:last_index], profile_1D[:last_index])

    @classmethod
    def __create_2D_profile_from_1D(cls, radius, profile_1D):
        interpol_index = interpolate.interp1d(radius, profile_1D, bounds_error=False, fill_value=0.0)

        xv = numpy.arange(-radius[-1], radius[-1], radius[1] - radius[0])  # adjust your matrix values here
        X, Y = numpy.meshgrid(xv, xv)
        profilegrid = numpy.zeros(X.shape, float)
        for i, x in enumerate(X[0, :]):
            for k, y in enumerate(Y[:, 0]):
                current_radius = numpy.sqrt(x ** 2 + y ** 2)
                profilegrid[i, k] = interpol_index(current_radius)

        return X, Y, profilegrid

    ###################################################
    #
    def __build_zone_plate_profile(self):
        at = self.__attributes
        op = self.__options

        radia = numpy.sqrt(numpy.arange(0, self.n_zones + 1) * self.wavelength * self.focal_distance + ((numpy.arange(0, self.n_zones + 1) * self.wavelength) ** 2) / 4)
        profile = numpy.full(self.n_points, 1 + 0j)
        profile[int(numpy.floor(radia[self.n_zones] / self.step)):self.n_points] = 0

        # Ordinary FZP
        if op.zone_plate_type == 0:
            for i in range(1, self.n_zones, 2):
                position_i = int(numpy.floor(radia[i] / self.step))
                position_f = int(numpy.floor(radia[i + 1] / self.step))  # N.B. the index is excluded
                profile[position_i:position_f] = numpy.exp(-1j * (-2 * numpy.pi * self.delta_FZP / self.wavelength * at.height - 1j * 2 * numpy.pi * self.beta_FZP / self.wavelength * at.height))

            membrane_transmission = 1

        # Zone-doubled FZP
        if op.zone_plate_type == 1:
            for i in range(1, self.n_zones, 2):
                position_i = int(numpy.floor((radia[i] + at.b_min / 4) / self.step))
                position_f = int(numpy.floor((radia[i + 1] - at.b_min / 4) / self.step))
                profile[position_i:position_f] = numpy.exp(-1j * (-2 * numpy.pi * self.delta_template / self.wavelength * at.height - 1j * 2 * numpy.pi * self.beta_template / self.wavelength * at.height))

                position_i = int(numpy.floor((radia[i] - op.width_coating / 2) / self.step))
                position_f = int(numpy.floor((radia[i] + op.width_coating / 2) / self.step))
                profile[position_i:position_f] = numpy.exp(-1j * (-2 * numpy.pi * self.delta_FZP / self.wavelength * at.height - 1j * 2 * numpy.pi * self.beta_FZP / self.wavelength * at.height))

                position_i = int(numpy.floor((radia[i + 1] - op.width_coating / 2) / self.step))
                position_f = int(numpy.floor((radia[i + 1] + op.width_coating / 2) / self.step))
                profile[position_i:position_f] = numpy.exp(-1j * (-2 * numpy.pi * self.delta_FZP / self.wavelength * at.height - 1j * 2 * numpy.pi * self.beta_FZP / self.wavelength * at.height))

            # including absorption of coating material 
            membrane_transmission = numpy.exp(-1j * (-1j * 2 * numpy.pi * self.beta_FZP / self.wavelength * op.width_coating / 2))

        # Zone-filled FZP
        if op.zone_plate_type == 2:
            for i in range(1, self.n_zones, 2):

                position_i = int(numpy.floor(radia[i] / self.step))
                position_f = int(numpy.floor(radia[i + 1] / self.step))

                width = numpy.abs(int(numpy.floor((radia[i + 1] - radia[i]) / self.step)))
                op.width_coating_step = numpy.abs(int(numpy.floor(op.width_coating / self.step / 2)))

                if op.width_coating < width:
                    profile[position_i:position_f] = numpy.exp(-1j * (-2 * numpy.pi * self.delta_template / self.wavelength * at.height - 1j * 2 * numpy.pi * self.beta_template / self.wavelength * at.height))

                    position_i = int(numpy.floor((radia[i] - op.width_coating) / self.step))
                    position_f = int(numpy.floor(radia[i] / self.step))
                    profile[position_i:position_f] = numpy.exp(-1j * (-2 * numpy.pi * self.delta_FZP / self.wavelength * at.height - 1j * 2 * numpy.pi * self.beta_FZP / self.wavelength * at.height))

                    position_i = int(numpy.floor(radia[i + 1] / self.step))
                    position_f = int(numpy.floor((radia[i + 1] + op.width_coating) / self.step))
                    profile[position_i:position_f] = numpy.exp(-1j * (-2 * numpy.pi * self.delta_FZP / self.wavelength * at.height - 1j * 2 * numpy.pi * self.beta_FZP / self.wavelength * at.height))
                else:
                    profile[position_i:position_f] = numpy.exp(-1j * (-2 * numpy.pi * self.delta_template / self.wavelength * at.height - 1j * 2 * numpy.pi * self.beta_template / self.wavelength * at.height))

                    position_i = int(numpy.floor((radia[i] - op.width_coating) / self.step))
                    position_f = int(numpy.floor(radia[i] / self.step))
                    profile[position_i:position_f] = numpy.exp(-1j * (-2 * numpy.pi * self.delta_FZP / self.wavelength * at.height - 1j * 2 * numpy.pi * self.beta_FZP / self.wavelength * at.height))

                    position_i = int(numpy.floor(radia[i + 1] / self.step))
                    position_f = int(numpy.floor((radia[i + 1] - op.width_coating) / self.step))
                    profile[position_i:position_f] = numpy.exp(-1j * (-2 * numpy.pi * self.delta_FZP / self.wavelength * at.height - 1j * 2 * numpy.pi * self.beta_FZP / self.wavelength * at.height))

            # including absorption of coating material 
            membrane_transmission = numpy.exp(-1j * (-1j * 2 * numpy.pi * self.beta_FZP / self.wavelength * op.width_coating))

        # Two-Level FZP - stop here refactoring
        if op.zone_plate_type == 3:
            height1 = op.height1_factor * at.height
            height2 = op.height2_factor * at.height

            for i in range(1, self.n_zones, 2):
                position_i = int(numpy.floor((2 * radia[i - 1] / 3 + radia[i + 1] / 3) / self.step))
                position_f = int(numpy.floor((radia[i - 1] / 3 + 2 * radia[i + 1] / 3) / self.step))
                profile[position_i:position_f] = numpy.exp(-1j * (-2 * numpy.pi * self.delta_FZP / self.wavelength * height1 - 1j * 2 * numpy.pi * self.beta_FZP / self.wavelength * height1))

                position_i = int(numpy.floor((radia[i - 1] / 3 + 2 * radia[i + 1] / 3) / self.step))
                position_f = int(numpy.floor((radia[i + 1]) / self.step))
                profile[position_i:position_f] = numpy.exp(-1j * (-2 * numpy.pi * self.delta_FZP / self.wavelength * height2 - 1j * 2 * numpy.pi * self.beta_FZP / self.wavelength * height2))

            membrane_transmission = 1

        # Inserting the CS
        # --------------------------------------------------------------------------

        if op.with_central_stop:
            cs_pix = numpy.floor(op.cs_diameter / self.step)
            profile[0: int(numpy.floor(cs_pix / 2))] = 0

        return profile, membrane_transmission

    ###################################################
    #
    def __get_profile_h(self, profile, r):
        # Recalculation of the position where the initial profile is defined.
        # Originally the profile is defined in position r0, that are linear for all
        # the values of position. Now we need to define the function in a new
        # coordinates that are by r. The next loop is interpolating the values of
        # the profile from the coordinates in r0 to the coordinates in r.
        # The output intensity profiles will be defined in r coordinates.
        r0 = numpy.arange(0, self.max_radius, self.step)
        profile_h = numpy.full(self.n_points, 0j)
        for i in range(0, self.n_points - 1):
            profile_h[i] = profile[i] + (profile[i + 1] - profile[i]) / (r0[i + 1] - r0[i]) * (r[i] - r0[i])
        profile_h[self.n_points - 1] = profile[self.n_points - 1]

        return profile_h

    ###################################################
    #
    def __propagate_multislicing(self, map_int, map_complex, field0, four0, q_max, q, c):
        step_slice = self.__attributes.height

        for n in range(self.n_slices-1):
            print("Propagation to slice #: ", n+2)

            proj = numpy.exp(-1j * step_slice * ((2 * numpy.pi * q) ** 2) / (2 * self.k))

            fun = numpy.multiply(proj, four0)
            field = hankel_transform(fun, q_max, c)
            fun = numpy.multiply(field0, field)

            map_int[1 + n, :] = numpy.multiply(numpy.abs(fun), numpy.abs(fun))
            map_complex[1 + n, :] = fun

            four0 = hankel_transform(fun, self.max_radius, c, self.n_zeros)

        return four0

    ###################################################
    #
    def __propagate_to_focus(self, map_int, map_complex, four0, q_max, q, c):
        op = self.__options

        if not op.with_order_sorting_aperture:
            self.__propagate_to_distance(map_int, map_complex, self.focal_distance, four0, q_max, q, c)
        else:
            # Propagation to OSA position and OSA insertion
            # --------------------------------------------------------------------------
            four_OSA = self.__propagate_to_OSA(four0, q_max, q, c)

            # Propagation at the focus position
            # --------------------------------------------------------------------------
            self.__propagate_to_distance(map_int, map_complex, self.focal_distance - op.osa_position, four_OSA, q_max, q, c)

    ###################################################
    #
    def __propagate_on_range(self, map_int, map_complex, four0, q_max, q, c):
        op = self.__options

        stepz = (op.range_f - op.range_i) / (op.n_z - 1)
        z = (op.range_i + numpy.arange(op.n_z) * stepz)

        if not op.with_order_sorting_aperture:
            for o in range(op.n_z):
                self.__propagate_to_distance(map_int, map_complex, z[o], four0, q_max, q, c, map_index=o)
        else:
            if op.osa_position < op.range_i:
                # Propagation to OSA position and OSA insertion
                # --------------------------------------------------------------------------
                four_OSA = self.__propagate_to_OSA(four0, q_max, q, c)

                # Continue the propagation from OSA on the range
                #--------------------------------------------------------------------------
                for o in range(op.n_z):
                    self.__propagate_to_distance(map_int, map_complex, z[o] - op.osa_position, four_OSA, q_max, q, c, map_index=o)
            else:
                z_before = z[numpy.where(z <= op.osa_position)]
                last_before = len(z_before)

                # Propagation from initial position to last position before OSA
                # ------------------------------------------------------------------
                for o in range(last_before):
                    self.__propagate_to_distance(map_int, map_complex, z_before[o], four0, q_max, q, c, map_index=o)

                # Propagation to OSA position and OSA insertion
                # --------------------------------------------------------------------------
                four_OSA = self.__propagate_to_OSA(four0, q_max, q, c)

                # Continue the propagation from first position after OSA to final position
                #--------------------------------------------------------------------------
                for o in range(last_before, op.n_z):
                    self.__propagate_to_distance(map_int, map_complex, z[o] - op.osa_position, four_OSA, q_max, q, c, map_index=o)

    ###################################################
    #
    def __propagate_to_distance(self, map_int, map_complex, z, four0, q_max, q, c, map_index=0):
        print("Propagation to distance: ", z, " m")

        proj = numpy.exp(-1j * z * ((2 * numpy.pi * q) ** 2) / (2 * self.k))
        fun = numpy.multiply(proj, four0)
        four11 = hankel_transform(fun, q_max, c)

        map_int[map_index + self.n_slices, :] = numpy.multiply(numpy.abs(four11), numpy.abs(four11))
        map_complex[map_index + self.n_slices, :] = four11

    ###################################################
    #
    def __propagate_to_OSA(self, four0, q_max, q, c):
        op = self.__options

        print("Propagation to OSA: ", op.osa_position, " m")
        # Propagation at the OSA position
        # --------------------------------------------------------------------------
        proj_OSA = numpy.exp(-1j * op.osa_position * ((2 * numpy.pi * q) ** 2) / (2 * self.k))
        fun = numpy.multiply(proj_OSA, four0)
        field_OSA = hankel_transform(fun, q_max, c)

        # Inserting OSA
        # --------------------------------------------------------------------------
        OSA_pix = int(numpy.floor(op.osa_diameter / self.step) - 1)
        field_OSA[int(OSA_pix / 2) + 1:self.n_points] = 0
        four_OSA = hankel_transform(field_OSA, self.max_radius, c)

        return four_OSA

    ###################################################
    #
    def __calculate_efficiency(self, map_index, map_out, profile_h, r, n_integration_points=0):
        shape = map_out.shape

        map_d = numpy.full((2 * shape[1], shape[0]), None)
        map_d[:self.n_points, :] = numpy.flipud(map_out[:, :self.n_points].T)
        map_d[self.n_points:2 * self.n_points, :] = map_out[:, :self.n_points].T

        if n_integration_points <= 0: n_integration_points = self.n_points

        I_dens_0 = numpy.zeros(self.n_points)
        I_dens_0[numpy.where(profile_h != 0.j)] = 1.0
        I_0 = numpy.trapz(numpy.multiply(r, I_dens_0), r)
        I = numpy.trapz(numpy.multiply(r[:n_integration_points + 1], map_d[self.n_points - 1:(self.n_points + n_integration_points), map_index]), r[:n_integration_points + 1])

        return numpy.divide(I, I_0)

from PyQt5.QtWidgets import QApplication, QWidget, QHBoxLayout

if __name__ == "__main__":
    app = QApplication([])

    with_range = False
    n_z = 5 if with_range else 1

    zs = FresnelZonePlateSimulator(options=FZPSimulatorOptions(with_range=with_range,
                                                               range_i=0.0160,
                                                               range_f=0.0162,
                                                               n_z=n_z,
                                                               with_multi_slicing=True,
                                                               n_slices=3),
                                   attributes=FZPAttributes())

    zs.initialize(energy_in_KeV=8.0, n_points=5000)

    map_int, _, efficiency = zs.simulate()

    print("Efficiency:", efficiency)

    container = QWidget()
    container.setFixedWidth(1500)

    layout = QHBoxLayout()

    figure = None
    for i in range(n_z):
        figure = zs.plot_1D(figure, map_int[zs.n_slices+i, :], 50, replace=i==0, profile_name="z pos #" + str(i+1))

    layout.addWidget(figure)
    layout.addWidget(zs.plot_2D(None, map_int[-1, :], 50))
    layout.addWidget(zs.plot_3D(None, map_int[zs.n_slices, :], 50))

    container.setLayout(layout)
    container.show()

    app.exec_()



# #########################################################################
# Copyright (c) 2020, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2020. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# #########################################################################
#% Calculates the Diffraction Efficiency of FZP
#% -------------------------------------------------------------------------
#% by Joan Vila, July 2010
#
#% Data from FZP simulations in 2D has to be loaded in memory.
#%
#% F_POS --> The position of the focus in units of Nz step has to be given.
#% N_points --> Number of points considered around the focal spot to
#%              integrate the intensity. (default = N)
#%
#% The code computes the incoming intensity by taking 1 for every pixel of
#% the structure. The intensity in focus is integrate considering the
#% profile that is obtained during the simulation.
#%
#% 27.11.2011 -- The formula for I had was taking the wrong range for map_d
#%               by one point. This has been corrected.
#%
#% 10.11.2010 -- The code was adapted to the output ot the codes that use
#%               Hankel Transform based on Manuel's code.
#%
#% 24.10.2010 -- The previous version was very roughly calculating the area
#%               under the intensity distribution with rough rectangle
#%               approximation. The new version is using a trapezoidal
#%               integration of MATLAB.
#% 01.09.2010 -- Version 1.0
#%
#% -------------------------------------------------------------------------

import numpy

def Efficiency_MGS(F_POS, map_d, profile_h, r, N, N_points=0):
    if N_points <= 0: N_points = N

    I_dens_0 = numpy.zeros(N)
    I_dens_0[numpy.where(profile_h != 0.j)] = 1.0
    #I_dens_0[numpy.where(numpy.logical_or(numpy.real(profile_h) != 0.0, numpy.imag(profile_h) != 0.0))] = 1.0
    I_0 = numpy.trapz(numpy.multiply(r, I_dens_0), r)
    I   = numpy.trapz(numpy.multiply(r[0:N_points+1], map_d[N-1:(N+N_points), F_POS]), r[0:N_points+1])

    return numpy.divide(I, I_0)



# #########################################################################
# Copyright (c) 2020, UChicago Argonne, LLC. All rights reserved.         #
#                                                                         #
# Copyright 2020. UChicago Argonne, LLC. This software was produced       #
# under U.S. Government contract DE-AC02-06CH11357 for Argonne National   #
# Laboratory (ANL), which is operated by UChicago Argonne, LLC for the    #
# U.S. Department of Energy. The U.S. Government has rights to use,       #
# reproduce, and distribute this software.  NEITHER THE GOVERNMENT NOR    #
# UChicago Argonne, LLC MAKES ANY WARRANTY, EXPRESS OR IMPLIED, OR        #
# ASSUMES ANY LIABILITY FOR THE USE OF THIS SOFTWARE.  If software is     #
# modified to produce derivative works, such modified software should     #
# be clearly marked, so as not to confuse it with the version available   #
# from ANL.                                                               #
#                                                                         #
# Additionally, redistribution and use in source and binary forms, with   #
# or without modification, are permitted provided that the following      #
# conditions are met:                                                     #
#                                                                         #
#     * Redistributions of source code must retain the above copyright    #
#       notice, this list of conditions and the following disclaimer.     #
#                                                                         #
#     * Redistributions in binary form must reproduce the above copyright #
#       notice, this list of conditions and the following disclaimer in   #
#       the documentation and/or other materials provided with the        #
#       distribution.                                                     #
#                                                                         #
#     * Neither the name of UChicago Argonne, LLC, Argonne National       #
#       Laboratory, ANL, the U.S. Government, nor the names of its        #
#       contributors may be used to endorse or promote products derived   #
#       from this software without specific prior written permission.     #
#                                                                         #
# THIS SOFTWARE IS PROVIDED BY UChicago Argonne, LLC AND CONTRIBUTORS     #
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT       #
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS       #
# FOR A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL UChicago     #
# Argonne, LLC OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT,        #
# INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING,    #
# BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;        #
# LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER        #
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT      #
# LIABILITY, OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN       #
# ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN IF ADVISED OF THE         #
# POSSIBILITY OF SUCH DAMAGE.                                             #
# #########################################################################
#%% Fresnel zone plate simulation code
#%--------------------------------------------------------------------------
#% by Joan Vila-Comamala from original IDL version of Ana Diaz (February, 2009)
#% June, 2010
#%
#% code modified by Michael Wojcik Oct, 2013
#%
#% It simulates wavefront after FZP and propagates to the focal plane
#% plots wavefield all through the propagation from the FZP to the focus.
#%
#% 2D Extension Code --> Wavefront propagation in made using the Hankel
#% transform for a circularly symmetric function a long the radial
#% coordinate. Hankel transform routine is Hankel_Transform_MGS.
#%
#% Keep your eyes open, this code has not been throughly debugged or tested!
#%
#%
#%% ------------------------------------------------------------------------

import numpy
from orangecontrib.shadow_advanced_tools.util.zone_plates.original import bessel_zeros
from orangecontrib.shadow_advanced_tools.util.zone_plates.original.hankel_transform_MGS import Hankel_Transform_MGS
from orangecontrib.shadow_advanced_tools.util.zone_plates.original.refractive_index import RefractiveIndex
from orangecontrib.shadow_advanced_tools.util.zone_plates.original.efficiency_MGS import Efficiency_MGS

## FZP, CS, OSA and simulation parameters
#--------------------------------------------------------------------------
  
with_CS = 1     # equal to 0 --> without CS - CENTRAL STOP
                # equal to 1 --> with CS
               
with_OSA = 0    # equal to 0 --> without OSA - ORDER SORTING APERTURE
                # equal to 1 --> with OSA
               
FZP_TYPE = 0    # equal to 0 --> Ordinary FZP
                # equal to 1 --> Zone-Doubled FZP
                # equal to 2 --> Zone-Filled FZP
                # equal to 3 --> Two-Level FZP
                # equal to 4 --> Three-Level FZP
                # equal to 5 --> ALD Multideposition FZP 1
                # equal to 6 --> ALD Multideposition FZP 2
                # equal to 7 --> Zone-Edge Slanted FZP
              
with_Range = 1  # equal to 0 --> plot to focal length
                # equal to 1 --> plot in a given range
               
with_MultiSlicing = 0 # equal to 1 --> apply multisilcing of the element 
                       # equal to 0 --> no multislicing

with_Complex = 0  # equal to 0 --> Complex Wavefront is not stored
                  # equal to 1 --> Complex Wavefront is storedk
 
energy = 8                         # photon energy [keV]
wavelength = 12.398/energy*1e-10   # wavelength [m]
k = 2*numpy.pi/wavelength          # wavevector [m-1]
height = 20000e-9                  # zone thickness or height [m]
diam = 50e-6                       # FZP diameter [m]
bmin = 50e-9                       # outermost zone width [m] / outermost period for ZD [m]
f = diam*bmin/wavelength           # focal distance [m]

Range_i = f-2e-6 # distance to FZP
Range_f = f+2e-6 # distance to FZP

CS_diam = 10e-6              # beamstop diameter [m]
OSA_position = 0.01          # distance FZP-OSA [m]
OSA_diam = 30e-6             # OSA diameter [m]

N = 5000                     # Number of sampling point in radial direction
R = diam                      # Radius of the simulation
step = R/N
Nzeros = int(numpy.floor(1.25*diam/2/R*N)) # Parameter to speed up the Hankel Transform
                                           # when the function has zeros for N > Nzero
Nz = 3          # Number of sampling points along the z axis
factor_z = 1.6  # Z axis range up to factor_z*f

#% FZP Material
#--------------------------------------------------------------------------

FZP_Material = 'Au'
Template_Material = 'SiO2'

delta_FZP, beta_FZP = RefractiveIndex(energy, FZP_Material)
delta_template, beta_template =  RefractiveIndex(energy, Template_Material)

if with_MultiSlicing == 1: NSlices = 100 # Number of slices of the element
else: NSlices = 1

width_coating = 20e-9                  # Width of the coated material for 
                                        # for a zone-filled FZP.
## Parameters for Stair Case Profiles 
#  Currently parameters for maximum at 6.2 keV using Au as FZP material                                        
                                       
#Two-level profile
height1 = (2/3)*1000e-9
height2 = (4/3)*1000e-9

## Contruction of the FZP profile
#--------------------------------------------------------------------------

Nzones = int(numpy.floor(1.0/4.0*(diam/bmin)))

radia = numpy.sqrt(numpy.arange(0, Nzones+1)*wavelength*f + ((numpy.arange(0, Nzones+1)*wavelength)**2)/4)
profile = numpy.full(N, 1 + 0j)
profile[int(numpy.floor(radia[Nzones]/step)):N] = 0

# Ordinary FZP
if FZP_TYPE == 0:
    for i in range (1, Nzones, 2):
       position_i = int(numpy.floor(radia[i]/step))
       position_f = int(numpy.floor(radia[i+1]/step)) # N.B. the index is excluded
       profile[position_i:position_f] = numpy.exp(-1j*(-2*numpy.pi*delta_FZP/wavelength*height-1j*2*numpy.pi*beta_FZP/wavelength*height))

    Membrane_Transmission = 1

# Zone-doubled FZP
if FZP_TYPE == 1:
    for i in range (1, Nzones, 2):
       position_i = int(numpy.floor((radia[i]+bmin/4)/step))
       position_f = int(numpy.floor((radia[i+1]-bmin/4)/step))
       profile[position_i:position_f] = numpy.exp(-1j*(-2*numpy.pi*delta_template/wavelength*height-1j*2*numpy.pi*beta_template/wavelength*height))
       
       position_i = int(numpy.floor((radia[i]-width_coating/2)/step))
       position_f = int(numpy.floor((radia[i]+width_coating/2)/step))
       profile[position_i:position_f] = numpy.exp(-1j*(-2*numpy.pi*delta_FZP/wavelength*height-1j*2*numpy.pi*beta_FZP/wavelength*height))
       
       position_i = int(numpy.floor((radia[i+1]-width_coating/2)/step))
       position_f = int(numpy.floor((radia[i+1]+width_coating/2)/step))
       profile[position_i:position_f] = numpy.exp(-1j*(-2*numpy.pi*delta_FZP/wavelength*height-1j*2*numpy.pi*beta_FZP/wavelength*height))

    #including absorption of coating material 
    Membrane_Transmission = numpy.exp(-1j*(-1j*2*numpy.pi*beta_FZP/wavelength*width_coating/2))

# Zone-filled FZP
if FZP_TYPE == 2:
    for i in range (1, Nzones, 2):
        
       position_i = int(numpy.floor(radia[i]/step))
       position_f = int(numpy.floor(radia[i+1]/step))
       
       width = numpy.abs(int(numpy.floor((radia[i+1]-radia[i])/step)))
       width_coating_step = numpy.abs(int(numpy.floor(width_coating/step/2)))
       
       if width_coating < width:
            profile[position_i:position_f] = numpy.exp(-1j*(-2*numpy.pi*delta_template/wavelength*height-1j*2*numpy.pi*beta_template/wavelength*height))
            
            position_i = int(numpy.floor((radia[i]-width_coating)/step))
            position_f = int(numpy.floor(radia[i]/step))
            profile[position_i:position_f] =  numpy.exp(-1j*(-2*numpy.pi*delta_FZP/wavelength*height-1j*2*numpy.pi*beta_FZP/wavelength*height))
            
            position_i = int(numpy.floor(radia[i+1]/step))
            position_f = int(numpy.floor((radia[i+1]+width_coating)/step))  
            profile[position_i:position_f] = numpy.exp(-1j*(-2*numpy.pi*delta_FZP/wavelength*height-1j*2*numpy.pi*beta_FZP/wavelength*height))
       else:
            profile[position_i:position_f] = numpy.exp(-1j*(-2*numpy.pi*delta_template/wavelength*height-1j*2*numpy.pi*beta_template/wavelength*height))
            
            position_i = int(numpy.floor((radia[i]-width_coating)/step))
            position_f = int(numpy.floor(radia[i]/step))
            profile[position_i:position_f] = numpy.exp(-1j*(-2*numpy.pi*delta_FZP/wavelength*height-1j*2*numpy.pi*beta_FZP/wavelength*height))
            
            position_i = int(numpy.floor(radia[i+1]/step))
            position_f = int(numpy.floor((radia[i+1]-width_coating)/step))
            profile[position_i:position_f] =  numpy.exp(-1j*(-2*numpy.pi*delta_FZP/wavelength*height-1j*2*numpy.pi*beta_FZP/wavelength*height))

    #including absorption of coating material 
    Membrane_Transmission = numpy.exp(-1j*(-1j*2*numpy.pi*beta_FZP/wavelength*width_coating))


# Two-Level FZP - stop here refactoring
if FZP_TYPE == 3:
    for i in range (1, Nzones, 2):
        position_i = int(numpy.floor((2*radia[i-1]/3+radia[i+1]/3)/step))
        position_f = int(numpy.floor((radia[i-1]/3+2*radia[i+1]/3)/step))
        profile[position_i:position_f] = numpy.exp(-1j*(-2*numpy.pi*delta_FZP/wavelength*height1-1j*2*numpy.pi*beta_FZP/wavelength*height1))
        
        position_i = int(numpy.floor((radia[i-1]/3+2*radia[i+1]/3)/step))
        position_f = int(numpy.floor((radia[i+1])/step))
        profile[position_i:position_f] = numpy.exp(-1j*(-2*numpy.pi*delta_FZP/wavelength*height2-1j*2*numpy.pi*beta_FZP/wavelength*height2))
   
    Membrane_Transmission = 1


# Inserting the CS
# --------------------------------------------------------------------------
CS_pix = numpy.floor(CS_diam / step)

if with_CS == 1: profile[0: int(numpy.floor(CS_pix / 2))] = 0

#% Propagation of the wavefield
# The routine performing the 0th order Hankel tranform is based in a
# algorithm that calculates the function at positision related to the zeros
# of the 0th order Bessel function.
#--------------------------------------------------------------------------

# Loading the position of the zeros, as much position as N+1. The file
# c.mat contains up to 200000 zeros of the 1st order Bessel function.
c = bessel_zeros['c'][0, 0:N+1]

# Definition of the position where the calculated input and transformated
# funtions are evaluated. We define also the maximum frequency in the
# angular domain.

Q = c[N]/(2*numpy.pi*R)   # Maximum frequency
r = c[0:N]*R/c[N]         # Radius vector
q = c[0:N]/(2*numpy.pi*R) # Frequency vector

## Recalculation of the position where the initial profile is defined.
# Originally the profile is defined in position r0, that are linear for all
# the values of position. Now we need to define the function in a new
# coordinates that are by r. The next loop is interpolating the values of
# the profile from the coordinates in r0 to the coordinates in r.
# The output intensity profiles will be defined in r coordinates.
#--------------------------------------------------------------------------

r0 = numpy.arange(0, R, step)
profile_h = numpy.full(N, 0j)
for i in range(0, N-1):
    profile_h[i] = profile[i]+(profile[i+1]-profile[i])/(r0[i+1]-r0[i])*(r[i]-r0[i])
profile_h[N-1] = profile[N-1]

## Calculation of the first angular spectrum
#--------------------------------------------------------------------------
map_int=numpy.zeros((NSlices+Nz, N))
if with_Complex == 1: map_complex = numpy.zeros((NSlices+Nz, N))

# The function 'Hankel_Transform_MGS' needs as input the field 'field0', 
# the maximum radius R and the zeros of the 0th Bessel function. In case of
# the inverse transformation from the angular domain back to spatial domain
# the function needs as a input the angular specturm 'fun', the maximum
# frequency Q and the zeros of the Bessel function.

field0 = profile_h*Membrane_Transmission
map_int[0, :] = numpy.multiply(numpy.abs(field0), numpy.abs(field0))
if with_Complex == 1: map_complex[0, :] = field0[0:N]

four0 = Hankel_Transform_MGS(field0, R, c)
field0 = profile_h

## Multi-Slicing of the FZP
#--------------------------------------------------------------------------
# Propagation of the wavefield inside the FZP is performed by slicing the
# element in pieces and considering free propagation between them. The
# volume effects in this way are considered.  
#

if with_MultiSlicing == 1:
    Step_Slice = height

    for n in range(NSlices-1):
        proj = numpy.exp(-1j*Step_Slice*((2*numpy.pi*q)**2)/(2*k))
        fun = numpy.multiply(proj, four0)
        field = Hankel_Transform_MGS(fun, Q, c)

        fun = numpy.multiply(field0, field)
        map_int[1+n, :] = numpy.multiply(numpy.abs(fun), numpy.abs(fun))
        if with_Complex == 1: map_complex[1+n, :] = fun

        four0 = Hankel_Transform_MGS(fun, R, c, Nzeros)

# Calculation from the FZP position to full z range
if with_Range == 0:
    stepz = factor_z*f/Nz
    z = numpy.arange(1, Nz+1)*stepz

    if with_OSA == 0:
        for o in range(Nz):
            print("Z position nr. ", o + 1, ": ",  z[o])
            proj = numpy.exp(-1j*z[o]*((2*numpy.pi*q)**2)/(2*k))
            fun = numpy.multiply(proj, four0)
            four11 = Hankel_Transform_MGS(fun, Q, c)
            map_int[o + NSlices, :] = numpy.multiply(numpy.abs(four11), numpy.abs(four11))
            if with_Complex == 1: map_complex[o + NSlices, :] = four11

    elif with_OSA == 1:
        OSA_pos = int(numpy.floor(OSA_position/stepz) - 1)

        for o in range(OSA_pos-1):
            print("(OSA) Z position nr. ", o + 1, ": ", z[o])
            proj = numpy.exp(-1j * z[o] * ((2 * numpy.pi * q)**2) / (2 * k))
            fun = numpy.multiply(proj, four0)
            four11 = Hankel_Transform_MGS(fun, Q, c)
            map_int[o + NSlices, :] = numpy.multiply(numpy.abs(four11), numpy.abs(four11))
            if with_Complex == 1: map_complex[o + NSlices, :] = four11

        # Propagation at the OSA position
        #--------------------------------------------------------------------------

        proj_OSA = numpy.exp(-1j * z[OSA_pos] * ((2 * numpy.pi * q)**2) / (2 * k))
        fun = numpy.multiply(proj_OSA, four0)
        field_OSA = Hankel_Transform_MGS(fun, Q, c)

        # Inserting OSA
        #--------------------------------------------------------------------------

        OSA_pix = int(numpy.floor(OSA_diam/step) - 1)
        field_OSA[int(OSA_pix/2)+1:N] = 0

        map_int[OSA_pos+NSlices,:] = numpy.multiply(numpy.abs(field_OSA), numpy.abs(field_OSA))
        four_OSA =  Hankel_Transform_MGS(field_OSA,R,c)
        if with_Complex == 1: map_complex[OSA_pos+NSlices,:] = field_OSA

        # Continue the propagation from OSA to focus
        #--------------------------------------------------------------------------
        for o in range(OSA_pos+1, Nz):
            print("(OSA) Z position nr. ", o + 1, ": ", z[o] - z[OSA_pos])
            proj = numpy.exp(-1j * (z[o] - z[OSA_pos]) * ((2 * numpy.pi * q)**2) / (2 * k))
            fun = numpy.multiply(proj, four_OSA)
            four11 = Hankel_Transform_MGS(fun, Q, c)
            map_int[o + NSlices, :] = numpy.multiply(numpy.abs(four11), numpy.abs(four11))
            if with_Complex == 1: map_complex[o + NSlices, :] = four11

elif with_Range == 1:
    stepz = (Range_f-Range_i)/(Nz-1)
    z = (Range_i + numpy.arange(Nz)*stepz)

    if (with_OSA == 0) or (OSA_position > Range_f):
        for o in range(Nz):
            print("Z position nr. ", o+1, ": ",  z[o])
            proj = numpy.exp(-1j*z[o]*((2*numpy.pi*q)**2)/(2*k))
            fun = numpy.multiply(proj, four0)
            four11 = Hankel_Transform_MGS(fun, Q, c)
            map_int[o + NSlices, :] = numpy.multiply(numpy.abs(four11), numpy.abs(four11))
            if with_Complex == 1: map_complex[o + NSlices, :] = four11
    elif with_OSA == 1:
        if OSA_position < Range_i:
            proj_OSA = numpy.exp(-1j*OSA_position*((2*numpy.pi*q)**2)/(2*k))
            fun = numpy.multiply(proj_OSA, four0)
            field_OSA = Hankel_Transform_MGS(fun, Q, c)
    
            # Inserting OSA
            #--------------------------------------------------------------------------
    
            OSA_pix = int(numpy.floor(OSA_diam/step) - 1)
            field_OSA[int(OSA_pix/2)+1:N] = 0
            four_OSA =  Hankel_Transform_MGS(field_OSA,R,c)
    
            for o in range(Nz):
                print("(OSA) Z position nr. ", o+1, ": ",  z[o] - OSA_position)
                proj = numpy.exp(-1j*(z[o] - OSA_position)*((2*numpy.pi*q)**2)/(2*k))
                fun = numpy.multiply(proj, four_OSA)
                four11 = Hankel_Transform_MGS(fun, Q, c)
                map_int[o + NSlices, :] = numpy.multiply(numpy.abs(four11), numpy.abs(four11))
                if with_Complex == 1: map_complex[o + NSlices, :] = four11
        else:
            # Propagation from Range_i to OSA
            #------------------------------------------------------------------
            OSA_pos =  int(numpy.floor((OSA_position-Range_i)/stepz) - 1)
    
            for o in range(OSA_pos-1):
                print("(OSA) Z position nr. ", o+1, ": ",  z[o])
                proj = numpy.exp(-1j*z[o]*((2*numpy.pi*q)**2)/(2*k))
                fun = numpy.multiply(proj, four0)
                four11 = Hankel_Transform_MGS(fun, Q, c)
                map_int[o + NSlices, :] = numpy.multiply(numpy.abs(four11), numpy.abs(four11))
                if with_Complex == 1: map_complex[o + NSlices, :] = four11

            # Propagation at the OSA position
            #--------------------------------------------------------------------------
    
            proj_OSA = numpy.exp(-1j * z[OSA_pos] * ((2 * numpy.pi * q)**2) / (2 * k))
            fun = numpy.multiply(proj_OSA, four0)
            field_OSA = Hankel_Transform_MGS(fun, Q, c)
    
            # Inserting OSA
            #--------------------------------------------------------------------------
    
            OSA_pix = int(numpy.floor(OSA_diam/step) - 1)
            field_OSA[int(OSA_pix/2)+1:N] = 0
    
            map_int[OSA_pos+NSlices,:] = numpy.multiply(numpy.abs(field_OSA), numpy.abs(field_OSA))
            four_OSA = Hankel_Transform_MGS(field_OSA,R,c)
            if with_Complex == 1: map_complex[OSA_pos+NSlices, :] = field_OSA
    
            # Continue the propagation from OSA to focus
            #--------------------------------------------------------------------------
            for o in range(OSA_pos+1, Nz):
                print("(OSA) Z position nr. ", o + 1, ": ", z[o] - z[OSA_pos])
                proj = numpy.exp(-1j * (z[o] - z[OSA_pos]) * ((2 * numpy.pi * q)**2) / (2 * k))
                fun = numpy.multiply(proj, four_OSA)
                four11 = Hankel_Transform_MGS(fun, Q, c)
                map_int[o + NSlices, :] = numpy.multiply(numpy.abs(four11), numpy.abs(four11))
                if with_Complex == 1: map_complex[o + NSlices, :] = four11

shape = map_int.shape

map = numpy.full((2*shape[1], shape[0]), None)

if with_Complex == 1:
    map[0:N,:]   = numpy.flipud(map_complex[:, 0:N].T)
    map[N:2*N,:] = map_complex[:, 0:N].T
else:
    map[0:N,:]   = numpy.flipud(map_int[:, 0:N].T)
    map[N:2*N,:] = map_int[:, 0:N].T


DE=Efficiency_MGS(3, map, profile_h, r, N, int(numpy.floor(10*bmin/step)))

print("Efficiency:", DE)

###################################################
#
# Plotting
#
###################################################
from scipy import interpolate

def rotate(r, profile):
    interpol_index = interpolate.interp1d(r, profile, bounds_error=False, fill_value=0.0)

    xv = numpy.arange(-r[-1], r[-1], r[1] - r[0])  # adjust your matrix values here
    X, Y = numpy.meshgrid(xv, xv)
    profilegrid = numpy.zeros(X.shape, float)
    for i, x in enumerate(X[0, :]):
        for k, y in enumerate(Y[:, 0]):
            current_radius = numpy.sqrt(x ** 2 + y ** 2)
            profilegrid[i, k] = interpol_index(current_radius)

    return profilegrid


from matplotlib import pyplot as plt

print("Shape:", map_int.shape)

for i in range(1, map_int.shape[0]):
    plt.plot(r0[:50], map_int[i, :50])
plt.show()

plt.imshow(rotate(r0[:50], map_int[1, :50]))
plt.show()
